import enum
import io
import pathlib
import traceback
from collections.abc import Coroutine
from dataclasses import dataclass
from typing import Any, cast, get_args, get_type_hints

import more_itertools
from pydantic import BaseModel
from pydantic_core import PydanticCustomError, PydanticUndefined
from pydantic_extra_types.color import Color, parse_str
from textual import on, work
from textual.app import App, ComposeResult
from textual.command import Hit, Hits, Provider
from textual.containers import (
    Horizontal,
    Right,
    Vertical,
    VerticalGroup,
    VerticalScroll,
)
from textual.message import Message
from textual.screen import Screen
from textual.validation import Integer, Number, ValidationResult, Validator
from textual.widgets import (
    Button,
    Collapsible,
    ContentSwitcher,
    Footer,
    Header,
    Input,
    Label,
    Link,
    ListView,
    Log,
    Markdown,
    MaskedInput,
    ProgressBar,
    Select,
    SelectionList,
    Static,
    Switch,
    Tab,
    TabbedContent,
    TabPane,
    Tabs,
)
from textual.widgets._list_item import ListItem
from textual.widgets.selection_list import Selection
from textual_fspicker import FileOpen, SelectDirectory
from upath import UPath

import libresvip
from libresvip.core.compat import ZipFile
from libresvip.core.config import DarkMode, Language, save_settings, settings
from libresvip.core.warning_types import CatchWarnings
from libresvip.extension.manager import get_translation, middleware_manager, plugin_manager
from libresvip.model.base import BaseComplexModel, Project
from libresvip.utils import translation
from libresvip.utils.text import supported_charset_names
from libresvip.utils.translation import gettext_lazy as _

translation.singleton_translation = get_translation()


class LibreSVIPCommandProvider(Provider):
    async def search(self, query: str) -> Hits:
        matcher = self.matcher(query)
        for name, help_text, callback, _discover in self.app.get_system_commands(self.screen):
            if (match := matcher.match(name)) > 0:
                yield Hit(
                    match,
                    matcher.highlight(name),
                    callback,
                    help=help_text,
                )


class PluginInfoScreen(Screen[None]):
    def __init__(self, plugin_id: str) -> None:
        self.plugin_id = plugin_id
        super().__init__()

    @on(Button.Pressed, "#close")
    def on_close(self, event: Button.Pressed) -> None:
        self.app.workers
        self.app.pop_screen()

    def compose(self) -> ComposeResult:
        yield Header(icon="☰")
        yield Footer()
        plugin_info = plugin_manager.plugin_registry[self.plugin_id]
        with Vertical():
            yield Label(plugin_info.name, classes="title")
            yield Label(f"{_('Version: ')}{plugin_info.version}")
            yield Link(f"{_('Author: ')}{_(plugin_info.author)}", url=plugin_info.website)
            yield Label(_("Introduction"))
            yield Markdown(_(plugin_info.description))
            with Horizontal():
                yield Label("", classes="fill-width")
                yield Button(_("Close"), id="close", variant="success")


class TaskLogScreen(Screen[None]):
    def __init__(self, log_text: str) -> None:
        self.log_text = log_text
        super().__init__()

    def on_mount(self) -> None:
        log = self.query_one(Log)
        log.write_line(self.log_text)

    @on(Button.Pressed, "#close")
    def on_close(self, event: Button.Pressed) -> None:
        self.app.pop_screen()

    @on(Button.Pressed, "#copy")
    def on_copy(self, event: Button.Pressed) -> None:
        self.app.copy_to_clipboard(self.log_text)
        self.app.notify(_("Copied"))

    def compose(self) -> ComposeResult:
        yield Header(icon="☰")
        yield Footer()
        with Vertical():
            yield Log()
            with Horizontal():
                yield Label("", classes="fill-width")
                yield Button(_("Copy to clipboard"), id="copy", variant="primary")
                yield Button(_("Close"), id="close", variant="success")


class SelectFormats(Vertical):
    @dataclass
    class InputFormatChanged(Message):
        value: str | None

    @dataclass
    class OutputFormatChanged(Message):
        value: str | None

    @on(Button.Pressed, "#swap_input_output")
    def on_swap_input_output(self, pressed: Button.Pressed) -> None:
        input_format_select = self.query_one("#input_format")
        output_format_select = self.query_one("#output_format")
        input_format_value, output_format_value = (
            output_format_select._value,
            input_format_select._value,
        )
        input_format_select._watch_value(input_format_value)
        output_format_select._watch_value(output_format_value)

    @on(Button.Pressed, "#input_plugin_info")
    def on_input_plugin_info(self, pressed: Button.Pressed) -> None:
        if settings.last_input_format is not None:
            self.app.push_screen(PluginInfoScreen(settings.last_input_format))

    @on(Button.Pressed, "#output_plugin_info")
    def on_output_plugin_info(self, pressed: Button.Pressed) -> None:
        if settings.last_output_format is not None:
            self.app.push_screen(PluginInfoScreen(settings.last_output_format))

    @on(Select.Changed, "#input_format")
    def on_input_format_changed(self, event: Select.Changed) -> None:
        last_input_format = event.value if event.value != Select.BLANK else None
        if settings.last_input_format != last_input_format:
            settings.last_input_format = last_input_format
            self.post_message(self.InputFormatChanged(value=last_input_format))

    @on(Select.Changed, "#output_format")
    def on_output_format_changed(self, event: Select.Changed) -> None:
        last_output_format = event.value if event.value != Select.BLANK else None
        if settings.last_output_format != last_output_format:
            settings.last_output_format = last_output_format
            self.post_message(self.OutputFormatChanged(value=last_output_format))

    def compose(self) -> ComposeResult:
        yield Label(_("Select File Formats"), classes="title")
        with Horizontal():
            yield Label(_("Import format"), classes="text-middle")
            yield Select(
                [
                    (f"{_(plugin_info.file_format)} (*.{plugin_info.suffix})", plugin_id)
                    for plugin_id, plugin_info in plugin_manager.plugin_registry.items()
                ],
                value=(
                    settings.last_input_format
                    if settings.last_input_format is not None
                    else next(
                        iter(plugin_manager.plugin_registry),
                        Select.BLANK,
                    )
                ),
                prompt="",
                id="input_format",
                classes="fill-width",
            )
            yield Button("ℹ", id="input_plugin_info", tooltip=_("View Detail Information"))
        with Horizontal():
            yield Label(_("Swap Input and Output"), classes="text-middle")
            yield Label("", classes="fill-width")
            yield Button("⇅", id="swap_input_output")
        with Horizontal():
            yield Label(_("Export format"), classes="text-middle")
            yield Select(
                [
                    (f"{_(plugin_info.file_format)} (*.{plugin_info.suffix})", plugin_id)
                    for plugin_id, plugin_info in plugin_manager.plugin_registry.items()
                ],
                value=(
                    settings.last_output_format
                    if settings.last_output_format is not None
                    else next(
                        iter(plugin_manager.plugin_registry),
                        Select.BLANK,
                    )
                ),
                prompt="",
                id="output_format",
                classes="fill-width",
            )
            yield Button("ℹ", id="output_plugin_info", tooltip=_("View Detail Information"))


class ColorValidator(Validator):
    def validate(self, value: str) -> ValidationResult:
        try:
            parse_str(value)
            return self.success()
        except PydanticCustomError:
            return self.failure(_("Invalid color format"))


class OptionsForm(ListView):
    def __init__(self, option_class: type[BaseModel] | None, *args: Any, **kwargs: Any) -> None:
        self.option_class = option_class
        self.option_dict = {} if option_class is None else option_class().model_dump(mode="json")
        super().__init__(*args, **kwargs)

    @on(Input.Changed)
    @on(Switch.Changed)
    @on(Select.Changed)
    def handle_value_changed(self, event: Input.Changed | Switch.Changed | Select.Changed) -> None:
        option_key = event.control.id.removeprefix("value_")
        self.option_dict[option_key] = event.value

    def compose(self) -> ComposeResult:
        if self.option_class is None:
            return
        for option_key, field_info in self.option_class.model_fields.items():
            default_value = None if field_info.default is PydanticUndefined else field_info.default
            with ListItem(), Horizontal():
                if option_key == "lyric_replacement_preset_name":
                    choices = [(preset, preset) for preset in settings.lyric_replace_rules]
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Select(
                        choices,
                        allow_blank=False,
                        value=default_value,
                        classes="fill-width",
                        id=f"value_{option_key}",
                    )
                elif option_key in ["encoding", "lyric_encoding"]:
                    choices = [(charset, charset) for charset in supported_charset_names()]
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Select(
                        choices,
                        allow_blank=False,
                        value=default_value,
                        classes="fill-width",
                        id=f"value_{option_key}",
                    )
                elif issubclass(field_info.annotation, enum.Enum):
                    default_value = str(default_value.value) if default_value else None
                    annotations = get_type_hints(
                        field_info.annotation,
                        include_extras=True,
                    )
                    choices = []
                    for enum_item in field_info.annotation:
                        if enum_item.name in annotations:
                            annotated_args = list(
                                get_args(annotations[enum_item.name]),
                            )
                            if len(annotated_args) >= 2:
                                enum_field = annotated_args[1]
                            else:
                                continue
                            choices.append((_(enum_field.title), str(enum_item.value)))
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Select(
                        choices,
                        allow_blank=False,
                        value=default_value,
                        classes="fill-width",
                        id=f"value_{option_key}",
                    )
                elif issubclass(field_info.annotation, bool):
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Label("", classes="fill-width")
                    yield Switch(default_value, id=f"value_{option_key}")
                elif issubclass(field_info.annotation, Color):
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Label("", classes="fill-width")
                    yield MaskedInput(
                        "#NNNnnn",
                        default_value,
                        id=f"value_{option_key}",
                        validators=[ColorValidator()],
                    )
                elif issubclass(field_info.annotation, int | float | str | BaseComplexModel):
                    if issubclass(field_info.annotation, BaseComplexModel):
                        default_value = field_info.annotation.default_repr()
                    elif not isinstance(default_value, str):
                        default_value = str(default_value)
                    validators = []
                    if issubclass(field_info.annotation, int):
                        validators.append(Integer())
                    elif issubclass(field_info.annotation, float):
                        validators.append(Number())
                    yield Label(_(field_info.title), classes="text-middle")
                    yield Input(
                        default_value,
                        id=f"value_{option_key}",
                        validators=validators,
                        classes="fill-width",
                    )
                else:
                    continue
                self.option_dict[option_key] = default_value
                if field_info.description:
                    yield Button("？", tooltip=_(field_info.description), disabled=True)


class TaskRow(Right):
    def __init__(
        self,
        input_path: pathlib.Path,
        stem: str,
        ext: str,
        arrow_symbol: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.input_path = input_path
        self.stem = stem
        self.ext = ext
        self.arrow_symbol = arrow_symbol
        self.log_text = ""
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Label(str(self.input_path), classes="text-middle half-width")
        yield Label(f"  {self.arrow_symbol}  ", classes="text-middle")
        if self.arrow_symbol != "⮥":
            yield Input(self.stem, id="stem", classes="fill-width")
            yield Label(self.ext, id="ext", classes="text-middle")
        else:
            yield Label("", classes="fill-width")
        yield Button("👁", id="view_log", tooltip=_("View Log"))

    @on(Input.Changed, "#stem")
    def handle_stem_change(self, event: Input.Changed) -> None:
        self.stem = event.value

    @on(Button.Pressed, "#view_log")
    def handle_view_log(self, event: Button.Pressed) -> None:
        if self.log_text:
            self.app.push_screen(TaskLogScreen(self.log_text))


class TUIApp(App[None]):
    TITLE = "LibreSVIP"
    COMMANDS = App.COMMANDS | {LibreSVIPCommandProvider}
    DEFAULT_CSS = (
        App.DEFAULT_CSS
        + """
    .title {
        width: 100%;
        text-align: center;
        text-style: bold;
    }
    .card {
        border: solid #666;
    }
    .text-middle {
        padding: 1 2;
    }
    .fill-width {
        width: 1fr;
    }
    .fill-height {
        height: 1fr;
    }
    .row {
        max-height: 5;
    }
    .top-pane {
        dock: top;
    }
    .bottom-pane {
        dock: bottom;
    }
    .task-row {
        layout: horizontal;
    }
    .half-width {
        width: 50%;
    }
    """
    )

    @property
    def dark(self) -> bool:
        return self.theme != "textual-light"

    @dark.setter
    def dark(self, value: bool) -> None:
        self.theme = "textual-dark" if value else "textual-light"

    def on_mount(self) -> None:
        self.temp_path = UPath("memory:/")
        theme_select = self.query_one("#theme_select")

        def update_theme(value: bool) -> None:
            if theme_select._value != value:
                with theme_select.prevent(Select.Changed):
                    theme_select._watch_value(int(value))

        theme_select.watch(self, "dark", update_theme, init=False)
        if settings.last_input_format is not None:
            self.post_message(SelectFormats.InputFormatChanged(settings.last_input_format))
        if settings.last_output_format is not None:
            self.post_message(SelectFormats.OutputFormatChanged(settings.last_output_format))

    def _on_exit_app(self) -> Coroutine[Any, Any, None]:
        save_settings()
        return super()._on_exit_app()

    def watch_dark(self, dark: bool) -> None:
        settings.dark_mode = DarkMode.DARK if dark else DarkMode.LIGHT
        return super().watch_dark(dark)

    @on(SelectFormats.InputFormatChanged)
    async def handle_input_format_change(self, event: SelectFormats.InputFormatChanged) -> None:
        if settings.reset_tasks_on_input_change:
            tab_id = self.query_one("#task_list").current
            self.query_one(f"ListView#{tab_id}").clear()
        if event.value:
            plugin_object = plugin_manager.plugin_registry[event.value].plugin_object
            if plugin_object is not None and (
                input_option := get_type_hints(plugin_object.load).get("options")
            ):
                input_options = self.query_one("#input_options")
                input_options.option_class = input_option
                input_options.option_dict = {}
                await input_options.recompose()

    @on(SelectFormats.OutputFormatChanged)
    async def handle_output_format_change(self, event: SelectFormats.OutputFormatChanged) -> None:
        if event.value:
            if settings.auto_set_output_extension:
                tab_id = self.query_one("#task_list").current
                task_list_view = self.query_one(f"ListView#{tab_id}")
                for node in task_list_view._nodes:
                    node.ext = event.value
                    node.query_one("#ext").update(f".{event.value}")
            plugin_object = plugin_manager.plugin_registry[event.value].plugin_object
            if plugin_object is not None and (
                output_option := get_type_hints(plugin_object.dump).get("options")
            ):
                output_options = self.query_one("#output_options")
                output_options.option_class = output_option
                output_options.option_dict = {}
                await output_options.recompose()

    @on(Button.Pressed, "#add_task")
    @work
    async def handle_add_task(self, event: Button.Pressed) -> None:
        if path_str := await self.push_screen_wait(
            FileOpen("/", title=_("Open"), open_button=_("Open"), cancel_button=_("Cancel")),
        ):
            selected_path = pathlib.Path(path_str)
        else:
            return
        ext = selected_path.suffix.removeprefix(".").lower()
        if ext in plugin_manager.plugin_registry and settings.auto_detect_input_format:
            settings.last_input_format = ext
            self.query_one("#input_format")._watch_value(ext)
        tab_id = self.query_one("#task_list").current
        task_list_view = self.query_one(f"ListView#{tab_id}")
        task_count = len(task_list_view)
        if tab_id == "direct" or task_count == 0:
            arrow_symbol = "→"
        elif tab_id == "split":
            return
        else:
            arrow_symbol = "⮥"
        task_list_view.append(
            ListItem(
                TaskRow(
                    selected_path,
                    selected_path.stem,
                    f".{settings.last_output_format}",
                    classes="task-row",
                    arrow_symbol=arrow_symbol,
                )
            )
        )

    @on(Switch.Changed)
    def handle_switch_changed(self, changed: Switch.Changed) -> None:
        if changed.switch.id == "auto_detect_import_format":
            settings.auto_detect_input_format = changed.value
        elif changed.switch.id == "reset_tasks_on_input_change":
            settings.reset_tasks_on_input_change = changed.value
        elif changed.switch.id == "auto_set_output_extension":
            settings.auto_set_output_extension = changed.value

    @on(Select.Changed, "#language_select")
    async def handle_language_changed(self, changed: Select.Changed) -> None:
        if settings.language.value != changed.value:
            settings.language = Language(changed.value)
            translation.singleton_translation = get_translation()
            await self.recompose()

    @on(Select.Changed, "#theme_select")
    def handle_theme_changed(self, changed: Select.Changed) -> None:
        if self.dark != changed.value:
            self.dark = bool(changed.value)

    @work(thread=True)
    def convert_one(
        self, progress_bar: ProgressBar, task_row_item: ListItem, *sub_task_items: list[ListItem]
    ) -> None:
        tab_id = self.query_one("#task_list").current
        task_row = task_row_item.get_child_by_type(TaskRow)
        sub_tasks = [
            cast("ListItem", sub_task_item).get_child_by_type(TaskRow)
            for sub_task_item in sub_task_items
        ]
        task_row.log_text = ""
        if settings.last_input_format is None or settings.last_output_format is None:
            return
        try:
            with CatchWarnings() as w:
                output_path = self.temp_path / task_row.stem
                if tab_id != "split":
                    output_path = output_path.with_suffix(
                        f".{settings.last_output_format}",
                    )
                input_plugin = plugin_manager.plugin_registry[settings.last_input_format]
                output_plugin = plugin_manager.plugin_registry[settings.last_output_format]
                if (
                    input_plugin.plugin_object is not None
                    and (
                        input_option_class := get_type_hints(input_plugin.plugin_object.load).get(
                            "options",
                        )
                    )
                    is not None
                    and output_plugin.plugin_object is not None
                    and (
                        output_option_class := get_type_hints(
                            output_plugin.plugin_object.dump,
                        ).get("options")
                    )
                    is not None
                ):
                    input_options = self.query_one("#input_options")
                    input_option = input_option_class.model_validate(input_options.option_dict)
                    if tab_id == "merge":
                        child_projects = [
                            input_plugin.plugin_object.load(
                                sub_task.input_path,
                                input_option,
                            )
                            for sub_task in more_itertools.value_chain(task_row, sub_tasks)
                        ]
                        project = Project.merge_projects(child_projects)
                    else:
                        project = input_plugin.plugin_object.load(
                            task_row.input_path,
                            input_option,
                        )
                    for (
                        middleware_id,
                        middleware,
                    ) in middleware_manager.plugin_registry.items():
                        enabled = self.query_one(f"#{middleware_id}_switch").value
                        if (
                            enabled
                            and middleware.plugin_object is not None
                            and hasattr(middleware.plugin_object, "process")
                        ):
                            option_form = self.query_one(f"#{middleware_id}_options")
                            project = middleware.plugin_object.process(
                                project,
                                option_form.option_class.model_validate(option_form.option_dict),
                            )
                    output_options = self.query_one("#output_options")
                    output_option = output_option_class.model_validate(output_options.option_dict)
                    if tab_id == "split":
                        output_path.mkdir(parents=True, exist_ok=True)
                        for i, child_project in enumerate(
                            project.split_tracks(settings.max_track_count)
                        ):
                            output_plugin.plugin_object.dump(
                                output_path
                                / f"{task_row.stem}_{i + 1:0=2d}.{settings.last_output_format}",
                                child_project,
                                output_option,
                            )
                    else:
                        output_plugin.plugin_object.dump(
                            output_path,
                            project,
                            output_option,
                        )
            if w.output:
                task_row.log_text = w.output
            buffer = io.BytesIO()
            if output_path.is_file():
                buffer.write(output_path.read_bytes())
            else:
                with ZipFile(buffer, "w") as zip_file:
                    for child_file in output_path.iterdir():
                        if not child_file.is_file():
                            continue
                        zip_file.writestr(
                            child_file.name,
                            child_file.read_bytes(),
                        )
            self.call_from_thread(
                self.deliver_text,
                io.StringIO(buffer.getvalue().decode("latin-1")),
                encoding="latin-1",
                save_directory=settings.save_folder,
                save_filename=output_path.name if tab_id != "split" else f"{task_row.stem}.zip",
            )
        except Exception:
            task_row.log_text = traceback.format_exc()
            self.call_from_thread(
                self.notify,
                f"Error occurred while converting {task_row.input_path}",
                severity="error",
            )
        self.call_from_thread(progress_bar.advance, 1)

    @work(thread=True, exclusive=True)
    def convert_all(self, task_list_view: ListView, progress_bar: ProgressBar) -> None:
        total = len(task_list_view) if task_list_view.id != "merge" else 1
        self.call_from_thread(progress_bar.update, total=total)
        if task_list_view.id == "merge":
            (task_row,), other_tasks = more_itertools.spy(task_list_view._nodes)
            self.convert_one(
                progress_bar,
                task_row,
                *other_tasks,
            )
        else:
            for task_row in task_list_view._nodes:
                self.convert_one(
                    progress_bar,
                    task_row,
                )

    @on(Button.Pressed, "#start_conversion")
    async def handle_start_conversion(self, event: Button.Pressed) -> None:
        tab_id = self.query_one("#task_list").current
        task_list_view = self.query_one(f"ListView#{tab_id}")
        if len(task_list_view):
            progress_bar = self.query_one(ProgressBar)
            self.convert_all(task_list_view, progress_bar)

    @on(Button.Pressed, "#delete_task")
    def handle_delete_task(self, event: Button.Pressed) -> None:
        tab_id = self.query_one("#task_list").current
        task_list_view = self.query_one(f"ListView#{tab_id}")
        item_len = len(task_list_view)
        if item_len > 0:
            task_list_view.pop(min(task_list_view.index or 0, item_len - 1))

    @on(Button.Pressed, "#clear_tasks")
    def handle_clear_tasks(self, event: Button.Pressed) -> None:
        tab_id = self.query_one("#task_list").current
        self.query_one(f"ListView#{tab_id}").clear()

    @on(Button.Pressed, "#change_output_directory")
    @work
    async def handle_change_output_directory(self, event: Button.Pressed) -> None:
        if directory := await self.push_screen_wait(
            SelectDirectory(
                "/",
                title=_("Select directory"),
                select_button=_("Select"),
                cancel_button=_("Cancel"),
            )
        ):
            settings.save_folder = pathlib.Path(directory)
            self.query_one("#output_directory").update(str(directory))

    @on(SelectionList.SelectedChanged)
    def handle_plugins_changed(self, selected: SelectionList.SelectedChanged) -> None:
        settings.disabled_plugins = list(
            set(selected.selection_list._values.keys())
            - set(selected.selection_list._selected.keys())
        )

    @on(Input.Changed, "#max_track_count")
    def handle_max_track_count_changed(self, event: Input.Changed) -> None:
        if event.validation_result.is_valid:
            settings.max_track_count = int(event.value)

    @on(Tabs.TabActivated, "#conversion_mode")
    def handle_conversion_mode_changed(self, activated: Tabs.TabActivated) -> None:
        task_list = self.query_one("#task_list")
        task_list.current = activated.tab.id
        max_track_count_input = self.query_one("#max_track_count")
        max_track_count_input.disabled = activated.tab.id != "split"

    def compose(self) -> ComposeResult:
        if settings.dark_mode == DarkMode.LIGHT:
            self.dark = False
        yield Header(icon="☰")
        yield Footer()
        with TabbedContent():
            with TabPane(_("Converter")), Horizontal():
                with Vertical():
                    yield SelectFormats(classes="card")
                    with Vertical(classes="card"):
                        yield Label(_("Conversion mode & Task list"), classes="title")
                        yield Tabs(
                            Tab(_("Direct"), id="direct"),
                            Tab(_("Split"), id="split"),
                            Tab(_("Merge"), id="merge"),
                            id="conversion_mode",
                        )
                        with VerticalScroll(), ContentSwitcher(initial="direct", id="task_list"):
                            yield ListView(id="direct")
                            yield ListView(id="split")
                            yield ListView(id="merge")
                with Vertical(classes="card"), VerticalScroll():
                    yield Label(_("Advanced Settings"), classes="title")
                    with Collapsible(title=_("Input Options")):
                        yield OptionsForm(None, id="input_options")
                    for middleware_id, middleware in middleware_manager.plugin_registry.items():
                        if middleware.plugin_object is not None and (
                            middleware_option := get_type_hints(
                                middleware.plugin_object.process
                            ).get("options")
                        ):
                            with Collapsible(title=_(middleware.name)), VerticalGroup():
                                with Horizontal(classes="row"):
                                    yield Label(_("Enable"), classes="text-middle fill-width")
                                    yield Switch(id=f"{middleware_id}_switch")
                                with Vertical():
                                    yield OptionsForm(
                                        middleware_option, id=f"{middleware_id}_options"
                                    )
                    with Collapsible(title=_("Output Options")):
                        yield OptionsForm(None, id="output_options")
                with Horizontal(classes="bottom-pane card row"):
                    yield Button("＋", tooltip=_("Add task"), id="add_task")
                    yield Button(
                        "▶", tooltip=_("Start conversion"), variant="primary", id="start_conversion"
                    )
                    yield Button("❌", tooltip=_("Delete selected task"), id="delete_task")
                    yield Button("🗑︎", tooltip=_("Clear tasks"), id="clear_tasks", variant="error")
                    yield Label(_("Progress"), classes="text-middle")
                    yield ProgressBar(total=100, show_eta=False, classes="text-middle fill-height")
                    yield Label(_("Max track count"), classes="text-middle")
                    yield Input(
                        value=str(settings.max_track_count),
                        id="max_track_count",
                        validators=[Integer(minimum=1)],
                        disabled=True,
                    )
            with TabPane(_("Basic Settings")), Vertical():
                with Vertical(classes="card"):
                    yield Label(_("Conversion Settings"), classes="title")
                    with Horizontal():
                        yield Label(_("Auto detect import format"), classes="text-middle")
                        yield Label("", classes="fill-width")
                        yield Switch(
                            value=settings.auto_detect_input_format, id="auto_detect_import_format"
                        )
                    with Horizontal():
                        yield Label(
                            _("Reset list when import format changed"), classes="text-middle"
                        )
                        yield Label("", classes="fill-width")
                        yield Switch(
                            value=settings.reset_tasks_on_input_change,
                            id="reset_tasks_on_input_change",
                        )
                with Vertical(classes="card"):
                    yield Label(_("Output Settings"), classes="title")
                    with Horizontal():
                        yield Label(
                            _("Set Output File Extension Automatically"), classes="text-middle"
                        )
                        yield Label("", classes="fill-width")
                        yield Switch(
                            value=settings.auto_set_output_extension, id="auto_set_output_extension"
                        )
                    with Horizontal():
                        yield Label(_("Output Folder"), classes="text-middle")
                        yield Static(
                            str(settings.save_folder),
                            classes="text-middle fill-width",
                            disabled=True,
                            id="output_directory",
                        )
                        yield Button(
                            "🗀", id="change_output_directory", tooltip=_("Change Output Directory")
                        )
            with TabPane(_("Other Settings")), Horizontal():
                with Vertical(classes="card"):
                    yield Label(_("Appearance"), classes="title")
                    with Horizontal():
                        yield Label(_("Switch Language"), classes="text-middle")
                        yield Select(
                            [
                                ("简体中文", "zh_CN"),
                                ("English", "en_US"),
                                ("Deutsch", "de_DE"),
                                # ("日本語", "ja_JP")
                            ],
                            value=settings.language.value,
                            prompt="",
                            allow_blank=False,
                            id="language_select",
                        )
                    with Horizontal():
                        yield Label(_("Switch Theme"), classes="text-middle")
                        yield Select(
                            [
                                (_("Light"), 0),
                                (_("Dark"), 1),
                            ],
                            value=int(self.dark),
                            prompt="",
                            allow_blank=False,
                            id="theme_select",
                        )
                with Vertical(classes="card"):
                    yield Label(_("Choose Plugins"), classes="title")
                    yield SelectionList(
                        *(
                            Selection(
                                f"{plugin_info.name} ({plugin_info.version})",
                                plugin_id,
                                plugin_id not in settings.disabled_plugins,
                            )
                            for plugin_id, plugin_info in plugin_manager.plugin_registry.items()
                        )
                    )
            with TabPane(_("About")), Vertical():
                yield Link(
                    "LibreSVIP 𝅘𝅥𝅮",
                    url="https://github.com/SoulMelody/LibreSVIP",
                    tooltip=_("Repo URL"),
                )
                yield Label(f"{_('Version: ')}{libresvip.__version__} 🔖")
                yield Link(
                    _("Author: SoulMelody") + " 🌐",
                    url="https://space.bilibili.com/175862486",
                    tooltip=_("Author's Profile"),
                )
                yield Label(_("Introduction") + " 📖")
                yield Markdown(f"""
{_("LibreSVIP is an open-sourced, liberal and extensionable framework that can convert your singing synthesis projects between different file formats.")}\n
{_("All people should have the right and freedom to choose. That's why we're committed to giving you a second chance to keep your creations free from the constraints of platforms and coterie.")}""")



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable


PointType = TypeVar("PointType")


class Point(NamedTuple):
    x: int
    y: int

    @classmethod
    def start_point(cls, value: int = -100) -> Point:
        return cls(-192000, value)

    @classmethod
    def end_point(cls, value: int = -100) -> Point:
        return cls(1073741823, value)


class PointList(abc.ABC, Generic[PointType]):
    root: list[PointType]

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> PointType:
        return self.root[index]

    def __setitem__(self, index: int, value: PointType) -> None:
        self.root[index] = value

    def __delitem__(self, index: int) -> None:
        del self.root[index]

    def __contains__(self, item: PointType) -> bool:
        return item in self.root

    def append(self, item: PointType) -> None:
        self.root.append(item)

    def insert(self, i: int, item: PointType) -> None:
        self.root.insert(i, item)

    def pop(self, i: int = -1) -> PointType:
        return self.root.pop(i)

    def remove(self, item: PointType) -> None:
        self.root.remove(item)

    def clear(self) -> None:
        self.root.clear()

    def count(self, item: PointType) -> int:
        return self.root.count(item)

    def index(self, item: PointType, *args: Any) -> int:
        return self.root.index(item, *args)

    def reverse(self) -> None:
        self.root.reverse()

    def sort(self, /, *args: Any, **kwds: Any) -> None:
        self.root.sort(*args, **kwds)

    def extend(self, other: PointList[PointType] | Iterable[PointType]) -> None:
        if isinstance(other, PointList):
            self.root.extend(other.root)
        else:
            self.root.extend(other)



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .dv_generator import DeepVocalGenerator
from .dv_parser import DeepVocalParser
from .model import dv_project_struct
from .options import InputOptions, OutputOptions


class DeepVocalConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        parser = DeepVocalParser(options)
        dv_project = dv_project_struct.parse(path.read_bytes())
        return parser.parse_project(dv_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        dv_project = DeepVocalGenerator(options).generate_project(project)
        dv_content = dv_project_struct.build(dv_project)
        path.write_bytes(dv_content)



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import SVProject
from .options import InputOptions, OutputOptions, SVProjectVersionCompatibility
from .synthv_generator import SynthVGenerator
from .synthv_parser import SynthVParser


class SynthVStudioConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        sv_content = (
            path.read_bytes().removesuffix(b"\x00").removeprefix("\ufeff".encode()).decode("utf-8")
        )
        sv_proj = SVProject.model_validate_json(sv_content, context={"path": path})
        if options.instant and sv_proj.instant_mode_enabled is not None:
            options.instant = sv_proj.instant_mode_enabled
        return SynthVParser(options=options).parse_project(sv_proj)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        sv_project = SynthVGenerator(
            options=options,
        ).generate_project(project)
        path.write_bytes(
            json.dumps(
                sv_project.model_dump(mode="json", by_alias=True, exclude_none=True),
                separators=(",", ":"),
            ).encode("utf-8")
            + (
                b""
                if options.version_compatibility == SVProjectVersionCompatibility.ABOVE_2_0_0
                else b"\x00"
            )
        )



import re
from collections.abc import Iterable
from importlib.resources import files

import jyutping
from bidict import bidict
from ko_pron.ko_pron import romanise

from libresvip.core.compat import json
from libresvip.core.constants import DEFAULT_PHONEME
from libresvip.core.lyric_phoneme.chinese import CHINESE_RE, WHITE_SPACE, get_pinyin_series
from libresvip.core.lyric_phoneme.japanese import to_romaji

from .constants import DEFAULT_DURATIONS, DEFAULT_PHONE_RATIO

resource_dir = files("libresvip.plugins.svp")
phoneme_categories_by_language = json.loads(
    (resource_dir / "phoneme_categories.json").read_text(encoding="utf-8")
)
phoneme_dictionary = {
    language: bidict(xsampa_dict) if language in ["mandarin", "cantonese"] else xsampa_dict
    for language, xsampa_dict in json.loads(
        (resource_dir / "phoneme_dictionary.json").read_text(encoding="utf-8")
    ).items()
}


def sv_g2p_one(buffer: list[str], language: str | None) -> Iterable[str]:
    if language == "japanese":
        graphemes = (to_romaji(part) for part in buffer)
        default_phoneme = "r a"
    elif language == "korean":
        graphemes = (romanise(part, "rr") for part in buffer)
        default_phoneme = "l 6"
    elif language == "mandarin":
        graphemes = get_pinyin_series(buffer, filter_non_chinese=False)  # type: ignore[assignment]
        default_phoneme = "l a"
    elif language == "cantonese":
        graphemes = (
            " ".join(re.sub(r"\d+$", " ", each) for each in jyutping.get(part)) for part in buffer
        )
        default_phoneme = "l a"
    else:
        yield from buffer
        return
    for grapheme in graphemes:
        yield phoneme_dictionary[language].get(grapheme, default_phoneme)


def sv_g2p(lyrics: Iterable[str], languages: Iterable[str]) -> list[str]:
    phoneme_list: list[str] = []
    builder: list[str] = []
    language = None
    for lyric, language in zip(lyrics, languages):
        if WHITE_SPACE.search(lyric) is not None and CHINESE_RE.search(lyric) is None:
            if builder:
                phoneme_list.extend(sv_g2p_one(builder, language))
                builder.clear()
            phoneme_list.append(lyric)
        else:
            builder.append(lyric)
    if builder:
        phoneme_list.extend(sv_g2p_one(builder, language))
    return phoneme_list


def xsampa2pinyin(xsampa: str, language: str) -> str:
    xsampa = re.sub(r"\s+", " ", xsampa).strip()
    return phoneme_dictionary[language].inverse.get(xsampa, DEFAULT_PHONEME)


def get_phoneme_categories(phoneme_parts: list[str], language: str) -> list[str]:
    if language in phoneme_categories_by_language:
        phoneme_categories = [
            phoneme_categories_by_language[language].get(phoneme) for phoneme in phoneme_parts
        ]
        if all(phoneme_category is not None for phoneme_category in phoneme_categories):
            return phoneme_categories
    return []


def number_of_phones(phoneme: str, language: str) -> int:
    phoneme_parts = phoneme.split()
    phoneme_categories = get_phoneme_categories(phoneme_parts, language)
    return len(phoneme_categories) if phoneme_categories else 2


def default_phone_marks(phoneme: str, language: str) -> list[float]:
    phoneme_parts = phoneme.split()
    res_len = max(len(phoneme_parts), 2)
    res = [0.0] * res_len
    if phoneme == "-":
        return res
    elif phoneme_categories := get_phoneme_categories(phoneme_parts, language):
        index = 0
        if phoneme_categories[index] in {"vowel", "diphthong"}:
            res[index] = getattr(DEFAULT_DURATIONS, phoneme_categories[index])
            index += 1
        if res_len > index:
            res[index:res_len] = [DEFAULT_PHONE_RATIO if index < res_len else 0.0] * (
                res_len - index
            )
    return res



from pydantic import Field

from libresvip.model.base import BaseModel


class UFNotes(BaseModel):
    key: int
    tick_on: int = Field(alias="tickOn")
    tick_off: int = Field(alias="tickOff")
    lyric: str


class UFPitch(BaseModel):
    ticks: list[int] = Field(default_factory=list)
    values: list[float | None] = Field(default_factory=list)
    is_absolute: bool = Field(alias="isAbsolute")


class UFTempos(BaseModel):
    tick_position: int = Field(alias="tickPosition")
    bpm: float


class UFTimeSignatures(BaseModel):
    measure_position: int = Field(alias="measurePosition")
    numerator: int = 4
    denominator: int = 4


class UFTracks(BaseModel):
    name: str
    notes: list[UFNotes] = Field(default_factory=list)
    pitch: UFPitch | None = None


class UFProject(BaseModel):
    name: str = "export"
    tracks: list[UFTracks] = Field(default_factory=list)
    time_signatures: list[UFTimeSignatures] = Field(default_factory=list, alias="timeSignatures")
    tempos: list[UFTempos] = Field(default_factory=list)
    measure_prefix: int = Field(alias="measurePrefix")


class UFData(BaseModel):
    format_version: int = Field(1, alias="formatVersion")
    project: UFProject



from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from libresvip.model.base import Note, SongTempo, TimeSignature


@dataclass
class KeyTick:
    tick: int
    tempo: SongTempo | None = None
    note_start: Note | None = None
    note_end: Note | None = None
    track_index: int | None = None


@dataclass
class MXmlMeasureContent:
    class NoteType(enum.IntEnum):
        BEGIN = 1
        MIDDLE = 2
        END = 3
        SINGLE = 4

    duration: int
    note: Note | None
    note_type: NoteType | None
    bpm: float | None = None

    @classmethod
    def with_tempo(cls, bpm: float) -> MXmlMeasureContent:
        return cls(duration=0, note=None, note_type=None, bpm=bpm)

    @classmethod
    def with_rest(cls, duration: int) -> MXmlMeasureContent:
        return cls(duration=duration, note=None, note_type=None, bpm=None)

    @classmethod
    def with_note(cls, duration: int, note: Note, note_type: NoteType) -> MXmlMeasureContent:
        return cls(duration=duration, note=note, note_type=note_type, bpm=None)


@dataclass
class MXmlMeasure:
    tick_start: int
    length: int
    time_signature: TimeSignature | None = None
    contents: list[MXmlMeasureContent] | None = None



from pydantic import BaseModel

from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    SelectSingleTrackMixin,
    StaticTempoMixin,
)


class InputOptions(EnablePitchImportationMixin, BaseModel):
    pass


class OutputOptions(SelectSingleTrackMixin, StaticTempoMixin):
    pass



from pydantic import Field

from libresvip.model.base import BaseModel


class VogenNote(BaseModel):
    pitch: int | None = None
    lyric: str | None = None
    rom: str | None = None
    on: int | None = None
    dur: int | None = None


class VogenTrack(BaseModel):
    name: str | None = None
    singer_id: str | None = Field(None, alias="singerId")
    rom_scheme: str | None = Field("", alias="romScheme")
    notes: list[VogenNote] = Field(default_factory=list)


class VogenProject(BaseModel):
    time_sig0: str = Field(alias="timeSig0")
    bpm0: float
    accom_offset: int | None = Field(0, alias="accomOffset")
    utts: list[VogenTrack] = Field(default_factory=list)



import enum
from collections.abc import ValuesView
from typing import get_args, get_type_hints

import typer
from rich.console import Console
from rich.table import Table

from libresvip.core.config import save_settings, settings
from libresvip.extension.manager import plugin_manager
from libresvip.extension.meta_info import FormatProviderPluginInfo
from libresvip.model.base import BaseComplexModel
from libresvip.utils.translation import gettext_lazy as _

app = typer.Typer()


@app.command()
def toggle(identifier: str) -> None:
    if identifier in plugin_manager.plugin_registry:
        try:
            settings.disabled_plugins.append(identifier)
            plugin_manager.import_plugins(reload=True)
            assert identifier not in plugin_manager.plugin_registry
            save_settings()
            typer.secho(_("The plugin is successfully disabled."), fg="green")
        except AssertionError:
            typer.secho(_("Failed to disable the plugin!"), err=True, fg="red")
    elif identifier in settings.disabled_plugins:
        try:
            settings.disabled_plugins.remove(identifier)
            plugin_manager.import_plugins(reload=True)
            assert identifier in plugin_manager.plugin_registry
            save_settings()
            typer.secho(_("The plugin is successfully enabled."), fg="green")
        except AssertionError:
            typer.secho(_("Failed to enable the plugin!"), err=True, fg="red")
    else:
        typer.secho(_("Unable to find the plugin."), fg="yellow")


@app.command("list")
def list_plugins() -> None:
    print_plugin_summary(plugin_manager.plugin_registry.values())


@app.command()
def detail(plugin_name: str) -> None:
    if plugin_name in plugin_manager.plugin_registry:
        print_plugin_details(plugin_manager.plugin_registry[plugin_name])
    else:
        typer.echo(_("Cannot find plugin ") + f"{plugin_name}!", err=True)


def print_plugin_summary(
    plugins: ValuesView[FormatProviderPluginInfo],
) -> None:
    console = Console(color_system="256")
    if not plugins:
        console.print(_("No plugins are currently installed."))
    margin = " " * 2
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column(_("No."), justify="left", style="cyan")
    table.add_column(_("Name"), justify="left", style="cyan")
    table.add_column(_("Version"), justify="left", style="cyan")
    table.add_column(_("Author"), justify="left", style="cyan")
    table.add_column(_("Identifier"), justify="left", style="cyan")
    table.add_column(_("Applicable file format"), justify="left", style="cyan")
    for num, plugin in enumerate(plugins):
        format_desc = f"{_(plugin.file_format)} (*.{plugin.suffix})"
        table.add_row(
            f"[{num + 1}] ",
            plugin.name + margin,
            str(plugin.version) + margin,
            _(plugin.author) + margin,
            plugin.suffix + margin,
            format_desc + margin,
        )
    console.print(table)


def print_plugin_details(plugin: FormatProviderPluginInfo) -> None:
    if plugin.plugin_object is None:
        return
    typer.echo()
    typer.echo("--------------------------------------------------\n")
    typer.echo(
        f"{{}}{plugin.name}\t{{}}{plugin.version!s}\t{{}}{_(plugin.author)}".format(
            _("Plugin: "),
            _("Version: "),
            _("Author: "),
        )
    )
    if plugin.website:
        typer.echo("\n" + _("Website: ") + plugin.website)
    format_desc = f"{_(plugin.file_format)} (*.{plugin.suffix})"
    typer.echo("\n" + f"{_('This plugin is applicable to')} {format_desc}.")
    typer.echo(
        _(
            "If you want to use this plugin, please specify '-i {}' (input) or '-o {}' (output) when converting."
        ).format(plugin.suffix.lower(), plugin.suffix.lower())
    )
    if plugin.description:
        typer.echo(f"\n{_('Description: ')}\n{_(plugin.description)}")
    op_arr = [_("input"), _("output")]
    options_arr = [
        get_type_hints(plugin.plugin_object.load).get("options", None),
        get_type_hints(plugin.plugin_object.dump).get("options", None),
    ]
    for op, options in zip(op_arr, options_arr):
        if options is None:
            continue
        typer.echo(_("This plugin supports the following {} conversion options:").format(op))
        for field_info in options.model_fields.values():
            if issubclass(field_info.annotation, bool | int | float | str | enum.Enum):
                typer.echo(
                    f"\n  {_(field_info.title)} = {field_info.annotation.__name__}    {_(field_info.description)}"
                )
                if field_info.default is not None:
                    typer.echo(f"\t{{}}{field_info.default}".format(_("Default: ")))
                if issubclass(field_info.annotation, enum.Enum):
                    typer.echo("  " + _("Available values:"))
                    annotations = get_type_hints(field_info.annotation, include_extras=True)
                    for enum_item in field_info.annotation:
                        if enum_item.name in annotations:
                            annotated_args = get_args(annotations[enum_item.name])
                            if len(annotated_args) == 2:
                                enum_type, enum_field = annotated_args
                                typer.echo(f"\t{enum_item.value}\t=>\t{_(enum_field.title)}")
            elif issubclass(field_info.annotation, BaseComplexModel):
                typer.echo(
                    f"\n  {_(field_info.title)} = {field_info.annotation.__name__}    {_(field_info.description)}"
                )
                typer.echo("  " + _("Available fields:"))
                for field in field_info.annotation.model_fields.values():
                    if hasattr(field.annotation, "__name__"):
                        typer.echo(f"    {field.title} = {field.annotation.__name__}")
                    else:
                        typer.echo(f"    {field.title} = {get_args(field.annotation)}")
                typer.echo("\t" + _("Default: ") + field_info.annotation.default_repr())
    typer.echo("\n--------------------------------------------------\n")



from typing import Final

MODE1_PITCH_SAMPLING_INTERVAL_TICK: Final[int] = 5
MODE2_PITCH_SAMPLING_INTERVAL_TICK: Final[int] = 4
MODE2_PITCH_MAX_POINT_COUNT: Final[int] = 50
MAX_ACCEPTED_BPM: Final[float] = 10000.0



from __future__ import annotations

import abc
import itertools
import sys
from types import SimpleNamespace
from typing import (
    Annotated,
    Literal,
    Protocol,
    runtime_checkable,
)

import more_itertools
from pydantic import BaseModel as PydanticBaseModel
from pydantic import (
    ConfigDict,
    Field,
    RootModel,
    ValidationInfo,
    computed_field,
    field_validator,
)

from libresvip.core.constants import DEFAULT_BPM, TICKS_IN_BEAT
from libresvip.model.point import Point, PointList


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
        # # Uncomment the following lines to enable strict mode
        # extra="forbid",
        # strict=True,
        # validate_assignment=True,
    )


@runtime_checkable
class BaseComplexModel(Protocol):
    @classmethod
    @abc.abstractmethod
    def default_repr(cls) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def from_str(cls, string: str) -> BaseComplexModel:
        pass


class Points(PointList[Point], RootModel[list[Point]]):
    root: list[Point] = Field(default_factory=list)


class SongTempo(BaseModel):
    position: int = Field(default=0, alias="Position")
    bpm: float = Field(default=DEFAULT_BPM, alias="BPM")


class TimeSignature(BaseModel):
    bar_index: int = Field(default=0, alias="BarIndex")
    numerator: int = Field(default=4, alias="Numerator")
    denominator: int = Field(default=4, alias="Denominator")

    def bar_length(self, ticks_in_beat: int = TICKS_IN_BEAT) -> float:
        return (ticks_in_beat * 4) * self.numerator / self.denominator


class ParamCurve(BaseModel):
    points: Points = Field(default_factory=Points, alias="PointList")

    @field_validator("points", mode="before")
    @classmethod
    def load_points(
        cls,
        points: Points | list[tuple[int, int]] | list[dict[str, int]],
        _info: ValidationInfo,
    ) -> Points:
        return (
            points if isinstance(points, Points) else Points(root=[Point(*each) for each in points])
        )

    @computed_field(alias="TotalPointsCount")
    def total_points_count(self) -> int:
        return len(self.points)

    def reduce_sample_rate(self, interval: int, interrupt_value: int = 0) -> ParamCurve:
        if interval <= 0:
            return self
        points = self.points
        if len(points) <= 1:
            return self
        result = []
        x_sum = 0
        y_sum = 0
        merged_count = 0
        prev_point = points[0]
        i = 0
        prev_point_added = False
        while i < len(points) - 1:
            i += 1
            prev_point_added = False
            current_point = points[i]
            if current_point[1] == interrupt_value:
                result.append(prev_point)
                prev_point = current_point
                continue
            elif prev_point[1] == interrupt_value:
                result.extend((prev_point, current_point))
                i += 1
                if i < len(points):
                    prev_point = points[i]
                prev_point_added = True
                continue
            pos = prev_point[0]
            first_loop = True
            while first_loop or (
                i < len(points) and points[i].x < pos + interval and points[i].y != interrupt_value
            ):
                if first_loop:
                    first_loop = False
                current_point = points[i]
                x_sum += prev_point.x
                y_sum += prev_point.y
                merged_count += 1
                prev_point = current_point
                i += 1
            result.append(
                Point(round(x_sum / merged_count), round(y_sum / merged_count)),
            )
            x_sum = 0
            y_sum = 0
            merged_count = 0
            if i >= len(points):
                break
            i -= 1
        if not prev_point_added:
            result.append(prev_point)
        return ParamCurve(points=Points(root=result))

    def split_into_segments(self, interrupt_value: int = 0) -> list[list[Point]]:
        segments: list[list[Point]] = []
        if len(self.points) == 0:
            return segments
        elif len(self.points) == 1:
            point = self.points[0]
            if point.x >= 0 and point.y != interrupt_value:
                segments.append([point])
            return segments
        buffer: list[Point] = []
        if interrupt_value != 0:
            for point in self.points.root:
                if point.x >= 0 and point.y < sys.maxsize // 2:
                    if point.y != interrupt_value:
                        buffer.append(point)
                    elif buffer:
                        segments.append(buffer.copy())
                        buffer.clear()
        else:
            current_point = self.points[0]
            i = 1
            while i < len(self.points):
                next_point = self.points[i]
                if current_point.y != interrupt_value:
                    buffer.append(current_point)
                elif next_point.y != interrupt_value:
                    if current_point.x >= 0 and (i <= 1 or self.points[i - 2].y != interrupt_value):
                        buffer.append(current_point)
                elif buffer:
                    segments.append(buffer.copy())
                    buffer.clear()
                current_point = next_point
                i += 1
            if current_point.x < sys.maxsize // 2 and (
                current_point.y != interrupt_value
                or i <= 1
                or self.points[i - 2].y != interrupt_value
            ):
                buffer.append(current_point)
        if buffer:
            segments.append(buffer.copy())
        return segments


class Params(BaseModel):
    pitch: ParamCurve = Field(default_factory=ParamCurve, alias="Pitch")
    volume: ParamCurve = Field(default_factory=ParamCurve, alias="Volume")
    breath: ParamCurve = Field(default_factory=ParamCurve, alias="Breath")
    gender: ParamCurve = Field(default_factory=ParamCurve, alias="Gender")
    strength: ParamCurve = Field(default_factory=ParamCurve, alias="Strength")


class VibratoParam(BaseModel):
    start_percent: float = Field(0.0, alias="StartPercent")
    end_percent: float = Field(0.0, alias="EndPercent")
    is_anti_phase: bool = Field(False, alias="IsAntiPhase")
    amplitude: ParamCurve = Field(default_factory=ParamCurve, alias="Amplitude")
    frequency: ParamCurve = Field(default_factory=ParamCurve, alias="Frequency")


class Phones(BaseModel):
    head_length_in_secs: float = Field(-1.0, alias="HeadLengthInSecs")
    mid_ratio_over_tail: float = Field(-1.0, alias="MidRatioOverTail")


class Note(BaseModel):
    start_pos: int = Field(0, alias="StartPos")
    length: int = Field(0, alias="Length")
    key_number: int = Field(0, alias="KeyNumber")
    head_tag: str | None = Field(None, alias="HeadTag")
    lyric: str = Field("", alias="Lyric")
    pronunciation: str | None = Field(None, alias="Pronunciation")
    edited_phones: Phones | None = Field(None, alias="EditedPhones")
    vibrato: VibratoParam | None = Field(None, alias="Vibrato")

    @property
    def end_pos(self) -> int:
        return self.start_pos + self.length


class TrackMixin(abc.ABC, BaseModel):
    title: str = Field("", alias="Title")
    mute: bool = Field(False, alias="Mute")
    solo: bool = Field(False, alias="Solo")
    volume: float = Field(1.0, alias="Volume")
    pan: float = Field(0.0, alias="Pan")


TrackType = SimpleNamespace(
    SINGING="Singing",
    INSTRUMENTAL="Instrumental",
)


class SingingTrack(TrackMixin):
    type_: Literal["Singing"] = Field(default=TrackType.SINGING, alias="Type")
    ai_singer_name: str = Field(default="", alias="AISingerName")
    reverb_preset: str = Field(default="", alias="ReverbPreset")
    note_list: list[Note] = Field(default_factory=list, alias="NoteList")
    edited_params: Params = Field(default_factory=Params, alias="EditedParams")


class InstrumentalTrack(TrackMixin):
    type_: Literal["Instrumental"] = Field(default=TrackType.INSTRUMENTAL, alias="Type")
    audio_file_path: str = Field(default="", alias="AudioFilePath")
    offset: int = Field(default=0, alias="Offset")


Track = Annotated[SingingTrack | InstrumentalTrack, Field(discriminator="type_")]


class Project(BaseModel):
    version: str = Field(default="", alias="Version")
    song_tempo_list: list[SongTempo] = Field(
        default_factory=list,
        alias="SongTempoList",
    )
    time_signature_list: list[TimeSignature] = Field(
        default_factory=list,
        alias="TimeSignatureList",
    )
    track_list: list[Track] = Field(default_factory=list, alias="TrackList")

    @classmethod
    def merge_projects(cls, projects: list[Project]) -> Project:
        assert len(projects) > 1, "No projects to merge"
        sample_project = projects[0]
        if not all(
            project.time_signature_list == sample_project.time_signature_list
            and project.song_tempo_list == sample_project.song_tempo_list
            for project in projects[1:]
        ):
            msg = "All projects must have the same time signatures and tempos"
            raise ValueError(msg)
        return sample_project.model_copy(
            update={
                "track_list": [
                    *itertools.chain.from_iterable(project.track_list for project in projects)
                ]
            },
            deep=True,
        )

    def split_tracks(self, max_track_count: int) -> list[Project]:
        assert any(isinstance(track, SingingTrack) for track in self.track_list), (
            "No singing tracks found"
        )
        return [
            self.model_copy(update={"track_list": track_chunk}, deep=True)
            for track_chunk in more_itertools.chunked(
                (track for track in self.track_list if isinstance(track, SingingTrack)),
                max_track_count,
            )
        ]



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .ace_studio_generator import AceGenerator
from .ace_studio_parser import AceParser
from .acep_io import compress_ace_studio_project, decompress_ace_studio_project
from .model import AcepProject
from .options import InputOptions, OutputOptions


class ACEStudioConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        obj = decompress_ace_studio_project(path)
        acep_project = AcepProject.model_validate(obj, context={"path": path})
        return AceParser(options=options).parse_project(acep_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ace_project = AceGenerator(options=options).generate_project(project)
        compress_ace_studio_project(ace_project.model_dump(mode="json", by_alias=True), path)



import pathlib

from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata_pydantic.bindings import XmlParser, XmlSerializer

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .cevio_generator import CeVIOGenerator
from .cevio_parser import CeVIOParser
from .model import CeVIOCreativeStudioProject
from .options import InputOptions, OutputOptions


class CeVIOConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        ccs_project = XmlParser(config=ParserConfig(fail_on_unknown_properties=False)).from_bytes(
            path.read_bytes(), CeVIOCreativeStudioProject
        )
        return CeVIOParser(options).parse_project(ccs_project)

    def dump(
        self,
        path: pathlib.Path,
        project: Project,
        options: OutputOptions,
    ) -> None:
        ccs_project = CeVIOGenerator(options).generate_project(project)
        serializer = XmlSerializer(
            config=SerializerConfig(pretty_print=True),
        )
        xml_text = serializer.render(ccs_project)
        path.write_bytes(xml_text.encode("utf-8"))



import dataclasses
import functools
import math
import operator
import pathlib
from typing import cast

import more_itertools
import portion
import sortedcontainers

from libresvip.core.tick_counter import shift_tempo_list
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.music_math import (
    db_to_float,
    linear_interpolation,
    ratio_to_db,
)

from .model import (
    S5pDbDefaults,
    S5pInstrumental,
    S5pMeterItem,
    S5pMixer,
    S5pNote,
    S5pParameters,
    S5pPoints,
    S5pProject,
    S5pTempoItem,
    S5pTrack,
)
from .options import InputOptions

TICK_RATE = 1470000


@dataclasses.dataclass
class SynthVEditorParser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    vibrato_value_interval_dict: PiecewiseIntervalDict = dataclasses.field(init=False)
    vibrato_coef_interval_dict: PiecewiseIntervalDict = dataclasses.field(init=False)

    def parse_project(self, s5p_project: S5pProject) -> Project:
        time_signature_list = self.parse_time_signatures(s5p_project.meter)
        self.first_bar_length = round(time_signature_list[0].bar_length())
        tempo_list = self.parse_tempos(s5p_project.tempo)
        self.synchronizer = TimeSynchronizer(tempo_list)
        track_list = self.parse_singing_tracks(s5p_project.tracks)
        if self.options.import_instrumental_track and s5p_project.instrumental.filename:
            track_list.append(
                self.parse_instrumental_track(s5p_project.instrumental, s5p_project.mixer)
            )
        return Project(
            time_signature_list=time_signature_list,
            song_tempo_list=tempo_list,
            track_list=track_list,
        )

    @staticmethod
    def parse_time_signatures(
        meter: list[S5pMeterItem],
    ) -> list[TimeSignature]:
        time_signatures = [
            TimeSignature(
                bar_index=item.measure,
                numerator=item.beat_per_measure,
                denominator=item.beat_granularity,
            )
            for item in meter
        ]
        if not time_signatures:
            time_signatures.append(TimeSignature(bar_index=0, numerator=4, denominator=4))
        return time_signatures

    def parse_tempos(self, tempo: list[S5pTempoItem]) -> list[SongTempo]:
        return shift_tempo_list(
            [
                SongTempo(
                    position=item.position // TICK_RATE,
                    bpm=item.beat_per_minute,
                )
                for item in tempo
            ],
            self.first_bar_length,
        )

    def parse_singing_tracks(self, tracks: list[S5pTrack]) -> list[Track]:
        return [
            SingingTrack(
                mute=track.mixer.muted,
                solo=track.mixer.solo,
                volume=self.parse_volume(track.mixer.gain_decibel),
                pan=track.mixer.pan,
                ai_singer_name=track.db_name or "",
                title=track.name or f"Track {i + 1}",
                note_list=note_list,
                edited_params=self.parse_params(track.parameters, note_list),
            )
            for i, track in enumerate(tracks)
            if (note_list := self.parse_notes(track.notes, track.db_defaults))
        ]

    @staticmethod
    def parse_volume(gain: float) -> float:
        if gain >= 0:
            return min(gain / (ratio_to_db(4)) + 1.0, 2.0)
        else:
            return db_to_float(gain)

    def parse_instrumental_track(
        self, track: S5pInstrumental, mixer: S5pMixer
    ) -> InstrumentalTrack:
        return InstrumentalTrack(
            mute=mixer.instrumental_muted,
            volume=self.parse_volume(mixer.gain_instrumental_decibel),
            title=pathlib.Path(track.filename).name,
            audio_file_path=track.filename,
            offset=round(self.synchronizer.get_actual_ticks_from_secs(track.offset)),
        )

    def parse_notes(
        self, s5p_notes: list[S5pNote | None], db_defaults: S5pDbDefaults
    ) -> list[Note]:
        self.vibrato_value_interval_dict = PiecewiseIntervalDict()
        self.vibrato_coef_interval_dict = PiecewiseIntervalDict()
        notes = []
        for s5p_note in s5p_notes:
            if s5p_note is None:
                continue
            note = Note(
                key_number=s5p_note.pitch,
                start_pos=round(s5p_note.onset / TICK_RATE),
                length=round(s5p_note.duration / TICK_RATE),
                lyric=s5p_note.lyric.lstrip(".").replace(" ", "") or db_defaults.lyric,
                pronunciation=s5p_note.comment
                if s5p_note.lyric.startswith(".") and s5p_note.comment
                else None,
            )
            notes.append(note)
            if self.options.import_pitch:
                vibrato_start = self.synchronizer.get_actual_secs_from_ticks(note.start_pos) + (
                    s5p_note.t_f0_vbr_start
                    if s5p_note.t_f0_vbr_start is not None
                    else db_defaults.t_f0_vbr_start
                )
                vibrato_end = self.synchronizer.get_actual_secs_from_ticks(note.end_pos)
                if vibrato_end <= vibrato_start:
                    continue
                phase = s5p_note.p_f0_vbr if s5p_note.p_f0_vbr is not None else db_defaults.p_f0_vbr
                frequency = (
                    s5p_note.f_f0_vbr if s5p_note.f_f0_vbr is not None else db_defaults.f_f0_vbr
                )
                if phase is not None and frequency is not None:
                    self.vibrato_value_interval_dict[
                        portion.closedopen(vibrato_start, vibrato_end)
                    ] = functools.partial(
                        self.s5p_vibrato_value,
                        vibrato_start=vibrato_start,
                        phase=phase,
                        frequency=frequency,
                    )
                vibrato_attack_time = vibrato_start + (
                    s5p_note.t_f0_vbr_left
                    if s5p_note.t_f0_vbr_left is not None
                    else db_defaults.t_f0_vbr_left
                )
                vibrato_release_time = vibrato_end - (
                    s5p_note.t_f0_vbr_right
                    if s5p_note.t_f0_vbr_right is not None
                    else db_defaults.t_f0_vbr_right
                )
                vibrato_depth = (
                    s5p_note.d_f0_vbr if s5p_note.d_f0_vbr is not None else db_defaults.d_f0_vbr
                )
                self.vibrato_coef_interval_dict[
                    portion.closedopen(vibrato_start, vibrato_attack_time)
                ] = functools.partial(
                    linear_interpolation,  # type: ignore[call-arg]
                    start=(vibrato_start, 0),
                    end=(vibrato_attack_time, vibrato_depth),
                )
                self.vibrato_coef_interval_dict[
                    portion.closedopen(vibrato_attack_time, vibrato_release_time)
                ] = vibrato_depth
                self.vibrato_coef_interval_dict[
                    portion.closedopen(vibrato_release_time, vibrato_end)
                ] = functools.partial(
                    linear_interpolation,  # type: ignore[call-arg]
                    start=(vibrato_release_time, vibrato_depth),
                    end=(vibrato_end, 0),
                )
        return notes

    @staticmethod
    def s5p_vibrato_value(
        seconds: float, vibrato_start: float, frequency: float, phase: float
    ) -> float:
        return math.sin(math.pi * (2 * (seconds - vibrato_start) * frequency + phase))

    def parse_params(self, parameters: S5pParameters, note_list: list[Note]) -> Params:
        params = Params()
        if self.options.import_pitch:
            rel_pitch_points = self.parse_pitch_curve(
                parameters.pitch_delta, parameters.interval, note_list
            )
            pitch_simulator = PitchSimulator(
                synchronizer=self.synchronizer,
                portamento=PortamentoPitch.sigmoid_portamento(),
                note_list=note_list,
            )
            params.pitch = RelativePitchCurve(self.first_bar_length).to_absolute(
                rel_pitch_points, pitch_simulator
            )
        return params

    def parse_pitch_curve(
        self, pitch_delta: S5pPoints, interval: int, note_list: list[Note]
    ) -> list[Point]:
        pitch_raw_data = sortedcontainers.SortedList(key=operator.attrgetter("x"))
        for s5p_point_group in more_itertools.split_when(
            pitch_delta.root, lambda p1, p2: p2.offset > p1.offset + 1
        ):
            for is_first, is_last, s5p_point in more_itertools.mark_ends(s5p_point_group):
                if is_first and is_last:
                    continue
                point_ticks = round(s5p_point.offset * (interval / TICK_RATE))
                point_value = s5p_point.value
                pitch_raw_data.add(
                    Point(
                        x=point_ticks,
                        y=round(point_value),
                    )
                )
        points = []
        prev_pos = None
        value = None
        for pos, value in pitch_raw_data:
            pos_secs = self.synchronizer.get_actual_secs_from_ticks(pos)
            if prev_pos is not None:
                for interp_pos in range(prev_pos + 5, pos, 5):
                    interp_pos_secs = self.synchronizer.get_actual_secs_from_ticks(interp_pos)
                    if interp_value_diff := self.vibrato_value_interval_dict.get(
                        interp_pos_secs, 0
                    ):
                        interp_value_diff *= self.vibrato_coef_interval_dict[interp_pos_secs] * 2000  # type: ignore[operator]
                    points.append(
                        Point(
                            x=interp_pos,
                            y=value + round(interp_value_diff),  # type: ignore[operator]
                        )
                    )
            if value_diff := self.vibrato_value_interval_dict.get(pos_secs, 0):
                value_diff *= cast("float", self.vibrato_coef_interval_dict[pos_secs]) * 2000
            points.append(
                Point(
                    x=pos,
                    y=value + round(value_diff or 0),
                )
            )
            prev_pos = pos
        return points



from libresvip.model.base import SongTempo, TimeSignature
from libresvip.utils.search import find_last_index


def skip_tempo_list(tempo_list: list[SongTempo], skip_ticks: int) -> list[SongTempo]:
    result = [
        tempo.model_copy(update={"position": tempo.position - skip_ticks})
        for tempo in tempo_list
        if tempo.position >= skip_ticks
    ]
    if not result or result[0].position > 0:
        i = find_last_index(tempo_list, lambda tempo: tempo.position <= skip_ticks)
        tempo = (
            SongTempo()
            if i == -1 and not tempo_list
            else tempo_list[i].model_copy(update={"position": 0})
        )
        result.insert(0, tempo)
    return result


def skip_beat_list(beat_list: list[TimeSignature], skip_bars: int) -> list[TimeSignature]:
    result = [
        beat.model_copy(update={"bar_index": beat.bar_index - skip_bars})
        for beat in beat_list
        if beat.bar_index >= skip_bars
    ]
    if not result or result[0].bar_index > 0:
        i = find_last_index(beat_list, lambda beat: beat.bar_index <= skip_bars)
        beat = (
            TimeSignature()
            if i == -1 and not beat_list
            else beat_list[i].model_copy(update={"bar_index": 0})
        )
        result.insert(0, beat)
    return result


def shift_tempo_list(tempo_list: list[SongTempo], shift_ticks: int) -> list[SongTempo]:
    result = tempo_list[:1]
    result.extend(
        tempo.model_copy(update={"position": tempo.position + shift_ticks})
        for tempo in tempo_list[1:]
    )
    return result


def shift_beat_list(beat_list: list[TimeSignature], shift_bars: int) -> list[TimeSignature]:
    result = beat_list[:1]
    result.extend(
        beat.model_copy(update={"bar_index": beat.bar_index + shift_bars}) for beat in beat_list[1:]
    )
    return result



from typing import Any

from parsimonious import Grammar, NodeVisitor
from parsimonious.nodes import Node
from pydantic import Field

from libresvip.model.base import BaseModel

nn_grammar = Grammar(
    r"""
    nn_project        =
        nn_info_line newline
        int newline
        (nn_note newline?)*
        nn_chord?
        nn_drum?

    nn_chord          =
        "#CHORD" newline
        int newline
        ('[' nn_legacy_points ']' newline)*

    nn_drum           =
        "#DRUM" newline
        '[' nn_legacy_points ']' newline

    newline           = ~"\r?\n"
    nn_time_signature = int " " int
    nn_info_line      = float ' ' nn_time_signature ' ' int ' ' int ' ' int ' 0 0 0 0'
    nn_points         = int (',' int)*
    nn_legacy_points  = int (', ' int)*
    nn_note           = ' ' word ' ' pinyin ' ' int ' ' int ' ' int ' ' int ' ' int ' ' int ' ' int ' ' int ' ' int ' ' nn_points ' ' nn_points ' ' int

    word              = ~"([\u4e00-\u9fff]|-|[a-z]+)"
    pinyin            = ~"[a-z-]+"
    float             = ~"[-+]?(\\d+(\\.\\d*)?|\\.\\d+)"
    int               = ~"[-+]?\\d+"
    """
)


class NNPoints(BaseModel):
    point_count: int = 100
    points: list[int] = Field(default_factory=lambda: [50] * 100)


class NNTimeSignature(BaseModel):
    numerator: int = 4
    denominator: int = 4


class NNInfoLine(BaseModel):
    tempo: float
    time_signature: NNTimeSignature
    bar_count: int = 0
    version: int = 19
    unknown: int = 0


class NNNote(BaseModel):
    lyric: str
    pronunciation: str
    start: int
    duration: int
    key: int
    cle: int = 0
    vel: int = 50
    por: int = 20
    vibrato_length: int = 0
    vibrato_depth: int = 0
    vibrato_rate: int = 0
    dynamics: NNPoints = Field(default_factory=NNPoints)
    pitch: NNPoints = Field(default_factory=NNPoints)
    pitch_bend_sensitivity: int = 12


class NNProject(BaseModel):
    info_line: NNInfoLine
    note_count: int = 0
    notes: list[NNNote] = Field(default_factory=list)


class NNVisitor(NodeVisitor):
    def visit_nn_project(self, node: Node, visited_children: list[Any]) -> NNProject:
        info_line, _, note_count, _, note_pairs, _, _ = visited_children
        return NNProject(
            info_line=info_line,
            note_count=note_count,
            notes=[note_pair[0] for note_pair in note_pairs],
        )

    def visit_nn_info_line(self, node: Node, visited_children: list[Any]) -> NNInfoLine:
        tempo, time_signature, bar_count, version, unknown = visited_children[::2]
        return NNInfoLine(
            tempo=tempo,
            time_signature=time_signature,
            bar_count=bar_count,
            version=version,
            unknown=unknown,
        )

    def visit_nn_time_signature(self, node: Node, visited_children: list[Any]) -> NNTimeSignature:
        numerator, _, denominator = visited_children
        return NNTimeSignature(numerator=numerator, denominator=denominator)

    def visit_nn_note(self, node: Node, visited_children: list[Any]) -> NNNote:
        (
            lyric,
            pronunciation,
            start,
            duration,
            key,
            cle,
            vel,
            por,
            vibrato_length,
            vibrato_depth,
            vibrato_rate,
            dynamics,
            pitch,
            pitch_bend_sensitivity,
        ) = visited_children[1::2]
        return NNNote(
            lyric=lyric,
            pronunciation=pronunciation,
            start=start,
            duration=duration,
            key=key,
            cle=cle,
            vel=vel,
            por=por,
            vibrato_length=vibrato_length,
            vibrato_depth=vibrato_depth,
            vibrato_rate=vibrato_rate,
            dynamics=dynamics,
            pitch=pitch,
            pitch_bend_sensitivity=pitch_bend_sensitivity,
        )

    def visit_nn_points(self, node: Node, visited_children: list[Any]) -> NNPoints:
        point_count, point_node_root = visited_children
        points = [point_node[-1] for point_node in point_node_root]
        return NNPoints(point_count=point_count, points=points)

    def visit_word(self, node: Node, visited_children: list[Any]) -> str:
        return node.text

    def visit_pinyin(self, node: Node, visited_children: list[Any]) -> str:
        return node.text

    def visit_int(self, node: Node, visited_children: list[Any]) -> int:
        return int(node.text)

    def visit_float(self, node: Node, visited_children: list[Any]) -> float:
        return float(node.text)

    def generic_visit(self, node: Node, visited_children: list[Any]) -> Any:
        return visited_children or node


nn_visitor = NNVisitor()



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VoiSonaProject, model_to_value_tree
from .options import InputOptions, OutputOptions
from .value_tree import JUCENode, build_tree_dict
from .voisona_generator import VoiSonaGenerator
from .voisona_parser import VoiSonaParser


class VoiSonaConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        value_tree = JUCENode.parse(path.read_bytes())
        tree_dict = build_tree_dict(value_tree)
        tssln_project = VoiSonaProject.model_validate(tree_dict["TSSolution"])
        return VoiSonaParser(options).parse_project(tssln_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        tssln_project = VoiSonaGenerator(options).generate_project(project)
        value_tree = model_to_value_tree(tssln_project)
        path.write_bytes(JUCENode.build(value_tree))



import dataclasses

from libresvip.core.lyric_phoneme.japanese import to_romaji
from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import (
    legato_chars,
    romaji2xsampa,
)
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.search import find_index

from .model import (
    PpsfAudioTrackEvent,
    PpsfAudioTrackItem,
    PpsfBaseSequence,
    PpsfCurvePoint,
    PpsfCurvePointSeq,
    PpsfCurveType,
    PpsfDvlTrackEvent,
    PpsfDvlTrackItem,
    PpsfEventTrack,
    PpsfFileAudioData,
    PpsfMeter,
    PpsfMeters,
    PpsfNote,
    PpsfParamPoint,
    PpsfProject,
    PpsfRegion,
    PpsfSeqParam,
    PpsfSyllable,
    PpsfTempo,
    PpsfTempos,
)
from .options import OutputOptions
from .ppsf_interval_dict import ppsf_key_interval_dict


@dataclasses.dataclass
class PiaproStudioGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> PpsfProject:
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        ppsf_project = PpsfProject()
        ppsf_project.ppsf.project.meter = self.generate_time_signatures(project.time_signature_list)
        ppsf_project.ppsf.project.tempo = self.generate_tempos(project.song_tempo_list)
        ppsf_project.ppsf.project.dvl_track = self.generate_singing_tracks(
            project.track_list,
            ppsf_project.ppsf.gui_settings.track_editor.event_tracks,
        )
        ppsf_project.ppsf.project.audio_track = self.generate_instrumental_tracks(
            project.track_list,
            ppsf_project.ppsf.gui_settings.track_editor.event_tracks,
        )
        ppsf_project.ppsf.gui_settings.project_length = self.first_bar_length
        for event_track in ppsf_project.ppsf.gui_settings.track_editor.event_tracks:
            for region in event_track.regions:
                ppsf_project.ppsf.gui_settings.project_length = max(
                    ppsf_project.ppsf.gui_settings.project_length,
                    region.position + region.length + self.first_bar_length,
                )
        return ppsf_project

    def generate_time_signatures(self, time_signatures: list[TimeSignature]) -> PpsfMeters:
        ppsf_meters = PpsfMeters()
        if len(time_signatures):
            ppsf_meters.const.nume = time_signatures[0].numerator
            ppsf_meters.const.denomi = time_signatures[0].denominator
        if len(time_signatures) > 1:
            ppsf_meters.use_sequence = True
            ppsf_meters.sequence = [
                PpsfMeter(
                    nume=time_signature.numerator,
                    denomi=time_signature.denominator,
                    measure=time_signature.bar_index,
                )
                for time_signature in time_signatures[1:]
            ]
        return ppsf_meters

    def generate_tempos(self, tempos: list[SongTempo]) -> PpsfTempos:
        ppsf_tempos = PpsfTempos()
        if len(tempos):
            ppsf_tempos.const = round(tempos[0].bpm * 10000)
        if len(tempos) > 1:
            ppsf_tempos.use_sequence = True
            ppsf_tempos.sequence = [
                PpsfTempo(
                    value=round(tempo.bpm * 10000),
                    tick=tempo.position,
                    curve_type=PpsfCurveType.BORDER,
                )
                for tempo in tempos[1:]
            ]
        return ppsf_tempos

    def generate_singing_tracks(
        self, tracks: list[Track], event_tracks: list[PpsfEventTrack]
    ) -> list[PpsfDvlTrackItem]:
        dvl_tracks = []
        for track in tracks:
            if isinstance(track, SingingTrack):
                if track.mute:
                    mute_flag = -1
                elif track.solo:
                    mute_flag = 1
                else:
                    mute_flag = 0
                event_track = PpsfEventTrack(
                    index=len(event_tracks),
                    track_type=4,
                    mute_solo=mute_flag,
                    notes=[],
                    nt_envelope_preset_id=2,
                )
                track_events, event_track.notes = self.generate_notes(track.note_list)
                event_track.regions.append(
                    PpsfRegion(
                        auto_expand_left=True,
                        auto_expand_right=True,
                        position=0,
                        length=(track_events[-1].pos + track_events[-1].length)
                        if len(track_events)
                        else 0,
                    )
                )
                key_interval_dict = ppsf_key_interval_dict(track_events, event_track.notes)
                pitch_param, pitch_curve_point = self.generate_pitch(
                    track.edited_params.pitch, key_interval_dict, track.note_list
                )
                event_track.curve_points.append(pitch_curve_point)
                dvl_track = PpsfDvlTrackItem(
                    name=track.title, events=track_events, parameters=[pitch_param]
                )
                dvl_tracks.append(dvl_track)
                event_tracks.append(event_track)
        return dvl_tracks

    def generate_pitch(
        self, pitch: ParamCurve, key_interval_dict: PiecewiseIntervalDict, notes: list[Note]
    ) -> tuple[PpsfSeqParam, PpsfCurvePoint]:
        pitch_param = PpsfSeqParam()
        curve_point = PpsfCurvePoint(
            sub_track_category=7,
            sub_track_id=2,
        )
        pitch_param.base_sequence = PpsfBaseSequence(
            constant=0,
            name="pitch_bend",
            use_sequence=True,
        )
        in_pitch_part = False
        for point in pitch.points.root:
            pos = point.x - self.first_bar_length
            if point.y == -100:
                if (
                    point.x not in [-192000, 1073741823]
                    and in_pitch_part
                    and len(pitch_param.base_sequence.sequence)
                ):
                    pitch_param.base_sequence.sequence.append(
                        PpsfParamPoint(
                            curve_type=PpsfCurveType.BORDER,
                            pos=pitch_param.base_sequence.sequence[-1].pos,
                            value=pitch_param.base_sequence.sequence[-1].value,
                        )
                    )
                    curve_point.sequence.append(
                        PpsfCurvePointSeq(
                            border_type=7,
                            abs_value=-1,
                            edited_by_user=True,
                            region_index=-1,
                            note_index=-1,
                            seg_array_id=-1,
                        )
                    )
                    in_pitch_part = False
            elif (base_key := key_interval_dict.get(pos)) is not None:
                if not in_pitch_part:
                    pitch_param.base_sequence.sequence.append(
                        PpsfParamPoint(
                            curve_type=PpsfCurveType.BORDER,
                            pos=pos - 1,
                            value=0,
                        )
                    )
                    curve_point.sequence.append(
                        PpsfCurvePointSeq(
                            border_type=6,
                            abs_value=-1,
                            edited_by_user=True,
                            region_index=-1,
                            note_index=-1,
                            seg_array_id=-1,
                        )
                    )
                    in_pitch_part = True
                pitch_param.base_sequence.sequence.append(
                    PpsfParamPoint(
                        curve_type=PpsfCurveType.NORMAL,
                        pos=pos,
                        value=point.y - round(base_key * 100),
                    )
                )
                curve_point.sequence.append(
                    PpsfCurvePointSeq(
                        border_type=1,
                        abs_value=point.y,
                        edited_by_user=True,
                        region_index=-1,
                        note_index=find_index(
                            notes, lambda note: note.start_pos <= pos <= note.end_pos
                        ),
                        seg_array_id=-1,
                    )
                )
        return pitch_param, curve_point

    def generate_instrumental_tracks(
        self, tracks: list[Track], event_tracks: list[PpsfEventTrack]
    ) -> list[PpsfAudioTrackItem]:
        audio_tracks = []
        for track in tracks:
            if (
                isinstance(track, InstrumentalTrack)
                and (track_info := audio_track_info(track.audio_file_path, only_wav=True))
                is not None
            ):
                offset = self.time_synchronizer.get_actual_secs_from_ticks(track.offset)
                tick_length = round(
                    self.time_synchronizer.get_actual_ticks_from_secs(
                        offset + track_info.duration / 1000
                    )
                    - track.offset
                )
                audio_track = PpsfAudioTrackItem(
                    name=track.title,
                    events=[
                        PpsfAudioTrackEvent(
                            tick_length=tick_length,
                            tick_pos=track.offset,
                            file_audio_data=PpsfFileAudioData(file_path=track.audio_file_path),
                        )
                    ],
                )
                audio_tracks.append(audio_track)
                if track.mute:
                    mute_flag = -1
                elif track.solo:
                    mute_flag = 1
                else:
                    mute_flag = 0
                event_tracks.append(
                    PpsfEventTrack(
                        index=len(event_tracks),
                        track_type=1,
                        mute_solo=mute_flag,
                        regions=[
                            PpsfRegion(
                                length=tick_length,
                                position=track.offset,
                                audio_event_index=0,
                            )
                        ],
                    )
                )
        return audio_tracks

    def generate_notes(self, notes: list[Note]) -> tuple[list[PpsfDvlTrackEvent], list[PpsfNote]]:
        track_events = []
        ppsf_notes = []
        for i, note in enumerate(notes):
            track_events.append(
                PpsfDvlTrackEvent(
                    note_number=note.key_number,
                    pos=note.start_pos,
                    length=note.length,
                    lyric=note.lyric,
                    symbols="-"
                    if note.lyric in legato_chars
                    else romaji2xsampa.get(to_romaji(note.lyric), "4 a"),
                )
            )
            ppsf_notes.append(
                PpsfNote(
                    region_index=0,
                    event_index=i,
                    length=note.length,
                    syllables=[
                        PpsfSyllable(
                            lyric_text=note.lyric,
                            symbols_text="-"
                            if note.lyric in legato_chars
                            else romaji2xsampa.get(to_romaji(note.lyric), "4 a"),
                        )
                    ],
                )
            )
        return track_events, ppsf_notes



# Ported from QNrbf by SineStriker
import dataclasses
import threading
from collections import ChainMap
from collections.abc import MutableMapping
from functools import cached_property
from types import TracebackType

from construct import Container
from typing_extensions import Self

from .binary_models import (
    classes_by_id,
    libraries_by_id,
    objects_by_id,
    references_by_id,
)


@dataclasses.dataclass
class NrbfIOBase:
    cur_thread_id: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.cur_thread_id = threading.get_ident()

    @cached_property
    def ref_map(self) -> MutableMapping[int, MutableMapping[int, Container]]:
        return ChainMap(
            classes_by_id[self.cur_thread_id],
            objects_by_id[self.cur_thread_id],
        )

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        classes_by_id.pop(self.cur_thread_id, None)
        objects_by_id.pop(self.cur_thread_id, None)
        libraries_by_id.pop(self.cur_thread_id, None)
        references_by_id.pop(self.cur_thread_id, None)



import enum
from typing import Any

from pydantic_core import PydanticUndefined
from pydantic_extra_types.color import Color
from rich.prompt import Confirm, FloatPrompt, IntPrompt, Prompt

from libresvip.core.config import get_ui_settings
from libresvip.model.base import BaseComplexModel, BaseModel
from libresvip.utils.text import supported_charset_names
from libresvip.utils.translation import gettext_lazy as _


def prompt_fields(option_class: BaseModel) -> dict[str, Any]:
    option_kwargs = {}
    if hasattr(option_class, "model_fields"):
        for i, (option_key, field_info) in enumerate(option_class.model_fields.items()):
            default_value = None if field_info.default is PydanticUndefined else field_info.default
            if field_info.title is None or field_info.annotation is None:
                continue
            translated_title = f"{i + 1}. {{}}".format(_(field_info.title))
            if option_key == "lyric_replacement_preset_name":
                choice = Prompt.ask(
                    translated_title,
                    choices=list(get_ui_settings().lyric_replace_rules),
                    default=default_value,
                )
                option_kwargs[option_key] = choice
            elif option_key in ["encoding", "lyric_encoding"]:
                choice = Prompt.ask(
                    translated_title,
                    choices=supported_charset_names(),
                    default=default_value,
                )
                option_kwargs[option_key] = choice
            elif issubclass(field_info.annotation, enum.Enum):
                default_value = str(default_value.value) if default_value else None
                choice = Prompt.ask(
                    translated_title,
                    choices=[str(x.value) for x in field_info.annotation],
                    default=default_value,
                )
                option_kwargs[option_key] = choice
            elif issubclass(field_info.annotation, bool):
                option_kwargs[option_key] = Confirm.ask(
                    translated_title,
                    default=default_value,
                )
            elif issubclass(field_info.annotation, int):
                option_kwargs[option_key] = IntPrompt.ask(
                    translated_title,
                    default=default_value,
                )
            elif issubclass(field_info.annotation, float):
                option_kwargs[option_key] = FloatPrompt.ask(
                    translated_title,
                    default=default_value,
                )
            elif issubclass(field_info.annotation, str | Color):
                option_kwargs[option_key] = Prompt.ask(
                    translated_title,
                    default=default_value,
                )
            elif issubclass(field_info.annotation, BaseComplexModel):
                value_str = Prompt.ask(
                    translated_title,
                    default=field_info.annotation.default_repr(),
                )
                option_kwargs[option_key] = field_info.annotation.from_str(value_str)
    return option_kwargs



# © 2018 Gerard Marull-Paretas <gerard@teslabs.com>
# © 2014 Mark Harviston <mark.harviston@gmail.com>
# © 2014 Arve Knudsen <arve.knudsen@gmail.com>
# BSD License

"""Mostly irrelevant, but useful utilities common to UNIX and Windows."""

import selectors

from PySide6 import QtCore

_fileno = selectors._fileobj_to_fd  # type: ignore[attr-defined]


def make_signaller(*types: type) -> QtCore.QObject:
    class Signaller(QtCore.QObject):
        signal = QtCore.Signal(*types)

    return Signaller()



import dataclasses

from libresvip.core.constants import DEFAULT_BPM
from libresvip.core.exceptions import NoTrackError
from libresvip.model.base import Project, SingingTrack, SongTempo
from libresvip.utils.translation import gettext_lazy as _

from .model import UTAUNote, UTAUProject, UTAUTrack
from .options import OutputOptions
from .pitch_mode1 import pitch_to_utau_mode1_track
from .pitch_mode2 import bpm_for_note, pitch_to_utau_mode2_track


@dataclasses.dataclass
class USTGenerator:
    options: OutputOptions

    def generate_project(self, project: Project) -> UTAUProject:
        if not project.song_tempo_list:
            project.song_tempo_list.append(SongTempo(bpm=DEFAULT_BPM))
        tempo = project.song_tempo_list[0].bpm
        if self.options.version < 2.0:
            if self.options.track_index < 0:
                first_singing_track = next(
                    (track for track in project.track_list if isinstance(track, SingingTrack)),
                    None,
                )
            else:
                first_singing_track = project.track_list[self.options.track_index]
            if first_singing_track is None or not isinstance(first_singing_track, SingingTrack):
                msg = _("No singing track found")
                raise NoTrackError(msg)
            ust_tracks = [self.generate_track(first_singing_track, project.song_tempo_list)]
        else:
            singing_tracks = [
                track for track in project.track_list if isinstance(track, SingingTrack)
            ]
            if not singing_tracks:
                msg = _("No singing track found")
                raise NoTrackError(msg)
            ust_tracks = [
                self.generate_track(track, project.song_tempo_list) for track in singing_tracks
            ]
        return UTAUProject(
            charset=self.options.encoding,
            ust_version=self.options.version,
            tempo=tempo,
            track=ust_tracks,
            pitch_mode2=True,
        )

    def generate_track(self, track: SingingTrack, tempo_list: list[SongTempo]) -> UTAUTrack:
        mode1_track_pitch_data = pitch_to_utau_mode1_track(
            track.edited_params.pitch, track.note_list
        )
        mode2_track_pitch_data = pitch_to_utau_mode2_track(
            track.edited_params.pitch, track.note_list, tempo_list
        )
        utau_notes = []
        prev_bpm = tempo_list[0].bpm
        prev_end_pos = None
        note_index = 0
        for note, mode1_pitch, mode2_pitch in zip(
            track.note_list,
            mode1_track_pitch_data.notes,
            mode2_track_pitch_data.notes,
        ):
            if rest_length := note.start_pos - (prev_end_pos if prev_end_pos is not None else 0):
                rest_note = UTAUNote(
                    note_type=str(note_index).zfill(4),
                    lyric="R",
                    length=rest_length,
                    note_num=60,
                )
                utau_notes.append(rest_note)
                note_index += 1
            cur_bpm = bpm_for_note(tempo_list, note)
            utau_note = UTAUNote(
                note_type=str(note_index).zfill(4),
                lyric=note.lyric,
                length=note.length,
                note_num=note.key_number,
                pitch_bend_points=mode1_pitch.pitch_points if mode1_pitch is not None else [],
                pitchbend_type="5",
                pitchbend_start=0,
            )
            if mode2_pitch:
                if mode2_pitch.start is not None:
                    utau_note.pbs = [mode2_pitch.start]
                if mode2_pitch.widths:
                    utau_note.pbw = [1, *mode2_pitch.widths]
                if mode2_pitch.start_shift is not None:
                    utau_note.pby = [mode2_pitch.start_shift]
                if mode2_pitch.shifts:
                    utau_note.pby += mode2_pitch.shifts
                if mode2_pitch.curve_types:
                    utau_note.pbm = mode2_pitch.curve_types
                if mode2_pitch.vibrato_params:
                    utau_note.vbr = mode2_pitch.vibrato_params
            if cur_bpm != prev_bpm:
                utau_note.tempo = cur_bpm
                prev_bpm = cur_bpm
            utau_notes.append(utau_note)
            prev_end_pos = note.end_pos
            note_index += 1
        return UTAUTrack(notes=utau_notes)



import dataclasses
import operator

import mido_fix as mido

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.lyric_phoneme.japanese import to_romaji
from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import (
    legato_chars,
    romaji2xsampa,
)
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.reset_time_axis import limit_bars

from .constants import DEFAULT_PHONEME
from .options import OutputOptions
from .vocaloid_pitch import generate_for_vocaloid


@dataclasses.dataclass
class VsqGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    @property
    def tick_rate(self) -> float:
        return TICKS_IN_BEAT / self.options.ticks_per_beat

    def generate_project(self, project: Project) -> mido.MidiFile:
        project = limit_bars(project, 4096)
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        mido_obj = mido.MidiFile(charset=self.options.lyric_encoding)
        mido_obj.ticks_per_beat = self.options.ticks_per_beat
        self.first_bar_length = int(
            project.time_signature_list[0].bar_length(self.options.ticks_per_beat)
        )
        master_track = mido.MidiTrack()
        master_track.name = "Master Track"
        self.generate_tempos(master_track, project.song_tempo_list)
        self.generate_time_signatures(master_track, project.time_signature_list)
        master_track.sort(key=operator.attrgetter("time"))
        mido_obj.tracks.append(master_track)
        mido_obj.tracks.extend(self.generate_tracks(project.track_list))
        self._convert_cumulative_to_delta(mido_obj.tracks)
        return mido_obj

    @staticmethod
    def _convert_cumulative_to_delta(tracks: list[mido.MidiTrack]) -> None:
        for track in tracks:
            tick = 0
            for event in track:
                tick, event.time = event.time, event.time - tick

    def generate_tempos(
        self, master_track: mido.MidiTrack, song_tempo_list: list[SongTempo]
    ) -> None:
        for tempo in song_tempo_list:
            master_track.append(
                mido.MetaMessage(
                    "set_tempo",
                    tempo=mido.bpm2tempo(tempo.bpm),
                    time=round(tempo.position / self.tick_rate),
                )
            )

    def generate_time_signatures(
        self,
        master_track: mido.MidiTrack,
        time_signature_list: list[TimeSignature],
    ) -> None:
        ticks = 0
        prev_time_signature = None
        for time_signature in time_signature_list:
            if prev_time_signature is not None:
                ticks += round(
                    prev_time_signature.bar_length(self.options.ticks_per_beat)
                    * (time_signature.bar_index - prev_time_signature.bar_index)
                )
            master_track.append(
                mido.MetaMessage(
                    "time_signature",
                    numerator=time_signature.numerator,
                    denominator=time_signature.denominator,
                    time=ticks,
                )
            )
            prev_time_signature = time_signature

    def generate_tracks(self, tracks: list[Track]) -> list[mido.MidiTrack]:
        mido_tracks = []
        singing_tracks = [
            track
            for track in tracks
            if isinstance(track, SingingTrack) and len(track.note_list) > 0
        ]
        for i, track in enumerate(singing_tracks):
            if (mido_track := self.generate_track(track, i, len(singing_tracks))) is not None:
                mido_tracks.append(mido_track)
        return mido_tracks

    def generate_track(
        self, track: SingingTrack, track_index: int, tracks_count: int
    ) -> mido.MidiTrack | None:
        track_text = self.generate_track_text(track, track_index, tracks_count)
        mido_track = mido.MidiTrack()
        while len(track_text) != 0:
            event_id = len(mido_track)
            event_id_str = str(event_id).zfill(4)
            header = f"DM:{event_id_str}:"
            available_length = 0x7F - len(header)
            mido_track.append(
                mido.MetaMessage(
                    "text",
                    text=header + track_text[:available_length],
                    time=0,
                )
            )
            track_text = track_text[available_length:]
        mido_track.append(mido.MetaMessage("end_of_track"))
        return mido_track

    def generate_track_text(
        self,
        track: SingingTrack,
        track_index: int,
        tracks_count: int,
        measure_prefix: int = 1,
    ) -> str:
        notes_lines = []
        lyrics_lines = []
        tick_prefix = self.first_bar_length
        tick_lists = [note.start_pos + tick_prefix for note in track.note_list]
        for i, note in enumerate(track.note_list):
            number = f"{i + 1}"
            notes_lines.extend(
                [
                    f"[ID#{number.zfill(4)}]",
                    "Type=Anote",
                    f"Length={note.length}",
                    f"Note#={note.key_number}",
                    "Dynamics=64",
                    "PMBendDepth=0",
                    "PMBendLength=0",
                    "PMbPortamentoUse=0",
                    "DEMdecGainRate=0",
                    "DEMaccent=0",
                    f"LyricHandle=h#{number.zfill(4)}",
                ]
            )
            lyric = note.lyric
            xsampa = (
                "-"
                if lyric in legato_chars
                else romaji2xsampa.get(to_romaji(lyric), DEFAULT_PHONEME)
            )
            lyrics_lines.extend(
                [
                    f"[h#{number.zfill(4)}]",
                    f"""L0="{lyric}","{xsampa}",0.000000,64,0,0""",
                ]
            )
        result = [
            "[Common]",
            "Version=DSB301",
            f"Name={track.title}",
            "Color=181,162,123",
            "DynamicsMode=1",
            "PlayMode=1",
        ]
        if track_index == 0:
            result.extend(
                [
                    "[Master]",
                    f"PreMeasure={measure_prefix}",
                    "[Mixer]",
                    "MasterFeder=0",
                    "MasterPanpot=0",
                    "MasterMute=0",
                    "OutputMode=0",
                    f"Tracks={tracks_count}",
                ]
            )
            for i in range(tracks_count):
                result.extend([f"Feder{i}=0", f"Panpot{i}=0", f"Mute{i}=0", f"Solo{i}=0"])
        result.extend(["[EventList]", "0=ID#0000"])
        result.extend(
            f"{tick}=ID#{str(index + 1).zfill(4)}" for index, tick in enumerate(tick_lists)
        )
        result.extend(
            [
                f"{track.note_list[-1].end_pos + tick_prefix}=EOS",
                "[ID#0000]",
                "Type=Singer",
                "IconHandle=h#0000",
            ]
        )
        result.extend(notes_lines)
        result.extend(
            [
                "[h#0000]",
                "IconID=$07010000",
                f"IDS={track.ai_singer_name}",
                "Original=0",
                "Caption=",
                "Length=1",
                "Language=0",
                "Program=0",
            ]
        )
        result.extend(lyrics_lines)
        result.extend(
            self.generate_pitch_text(track.edited_params.pitch, tick_prefix, track.note_list)
        )
        return "\n".join(result)

    def generate_pitch_text(
        self, pitch: ParamCurve, tick_prefix: int, note_list: list[Note]
    ) -> list[str]:
        result = []
        if pitch_raw_data := generate_for_vocaloid(
            pitch, note_list, self.first_bar_length, self.synchronizer
        ):
            if len(pitch_raw_data.pit):
                result.append("[PitchBendBPList]")
                result.extend(f"{pit.pos + tick_prefix}={pit.value}" for pit in pitch_raw_data.pit)
            if len(pitch_raw_data.pbs):
                result.append("[PitchBendSensBPList]")
                result.extend(f"{pbs.pos + tick_prefix}={pbs.value}" for pbs in pitch_raw_data.pbs)
        return result



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    SelectSingleTrackMixin,
    StaticTempoMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(EnablePitchImportationMixin, StaticTempoMixin, BaseModel):
    pass


class OutputOptions(SelectSingleTrackMixin, BaseModel):
    dict_name: str = Field(default="opencpop-extension", title=_("Dictionary Name"))
    split_threshold: float = Field(
        default=5,
        title=_("Split threshold (in seconds)"),
        description=_(
            "This option controls the segmentation strategy during conversion. When the value of this option is negative, no segmentation is performed; when the value of this option is 0, segmentation is performed at the threshold where the interval between all notes reaches the threshold; when the value of this option is positive, the minimum length of each segment can be controlled on the basis of segmentation. Setting a reasonable segmentation strategy can reduce the memory usage during synthesis while maximizing the utilization of performance and improving the synthesis effect."
        ),
    )
    min_interval: int = Field(
        default=400,
        title=_("Minimum interval (in milliseconds)"),
        description=_(
            "This option controls the minimum interval between notes. It is recommended to set it to a value greater than 300 milliseconds."
        ),
    )
    seed: int = Field(
        default=-1,
        title=_("Seed"),
        description=_(
            "A fixed random seed can get a stable and reproducible synthesis effect. This option takes effect when the non-negative value is set."
        ),
    )
    export_gender: bool = Field(
        default=False,
        title=_("Export gender parameter"),
    )
    indent: int = Field(
        default=2,
        title=_("Indentation"),
        description=_(
            "The number of spaces used for indentation. When the value is negative, no formatting is performed."
        ),
    )



from typing import Any

try:
    from lxml.etree import CDATA
    from xsdata.formats.dataclass.serializers.writers.lxml import (
        LxmlEventWriter,
    )

    class DefaultXmlWriter(LxmlEventWriter):
        def set_cdata(self, data: Any) -> None:
            self.flush_start(False)
            self.handler._element_stack[-1].text = CDATA(data=data)

except ImportError:
    from xml.sax.saxutils import XMLGenerator

    from xsdata.formats.dataclass.serializers.writers.native import (
        XmlEventWriter,
    )

    from .native import EchoGenerator

    class DefaultXmlWriter(XmlEventWriter):  # type: ignore[no-redef]
        def build_handler(self) -> XMLGenerator:
            return EchoGenerator(
                out=self.output,
                encoding=self.config.encoding,
                short_empty_elements=True,
            )

        def set_cdata(self, data: Any) -> None:
            self.flush_start(False)
            self.handler.start_cdata()
            super().set_data(data)
            self.handler.end_cdata()



import pathlib
import re
from typing import Any

from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata_pydantic.bindings import XmlParser, XmlSerializer

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.xmlutils import DefaultXmlWriter

from .model import VocalSharpProject
from .options import InputOptions, OutputOptions
from .vspx_generator import VocalSharpGenerator
from .vspx_parser import VocalSharpParser


class VocalSharpXMLWriter(DefaultXmlWriter):
    def set_data(self, data: Any) -> None:
        if (
            isinstance(data, str)
            and self.pending_tag
            and len(self.pending_tag) > 1
            and self.pending_tag[1]
            in (
                "name",
                "Name",
                "Singer",
                "lyric",
                "LSD",
                "symbol",
                "sample",
                "path",
            )
        ):
            super().set_cdata(data)
        else:
            super().set_data(data)

    def start_document(self) -> None:
        if self.config.xml_declaration:
            self.output.write(
                f'<?xml version="{self.config.xml_version}" encoding="{self.config.encoding}" standalone="no"?>\n'
            )


def strip_whitespace(matcher: re.Match[str]) -> str:
    first_tag = matcher[1]
    second_tag = matcher[2]
    if first_tag != second_tag:
        return f"</{first_tag}><{second_tag}>"
    else:
        return matcher.group()


def replace_self_closed(matcher: re.Match[str]) -> str:
    indent = matcher[1]
    tag = matcher[2]
    return f"{indent}<{tag}>{indent}</{tag}>"


class VocalSharpConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        vspx_project = XmlParser().from_bytes(path.read_bytes(), VocalSharpProject)
        return VocalSharpParser(options).parse_project(vspx_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        vspx_project = VocalSharpGenerator(options).generate_project(project)
        serializer = XmlSerializer(
            config=SerializerConfig(pretty_print=True, pretty_print_indent="\t"),
            writer=VocalSharpXMLWriter,
        )
        xml_text = serializer.render(vspx_project)
        xml_text = re.sub(r"(<[a-zA-Z]>)\s+", r"\1", xml_text)
        xml_text = re.sub(r"\s+(</[a-zA-Z]>)", r"\1", xml_text)
        xml_text = re.sub(r"</([a-z])>\s+<([a-z])>", strip_whitespace, xml_text)
        xml_text = re.sub(r"(\n\s*)<([a-zA-Z]+)/>", replace_self_closed, xml_text)
        path.write_bytes(xml_text.encode("utf-8"))



import dataclasses

from libresvip.core.tick_counter import shift_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import (
    PIT,
    VocalSharpBeat,
    VocalSharpDefaultTrill,
    VocalSharpMonoTrack,
    VocalSharpNote,
    VocalSharpNoteTrack,
    VocalSharpProject,
    VocalSharpStereoTrack,
    VocalSharpTempo,
)
from .options import InputOptions
from .vspx_interval_dict import BasePitchCurve


@dataclasses.dataclass
class VocalSharpParser:
    options: InputOptions
    default_trill: VocalSharpDefaultTrill | None = None
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, vspx_project: VocalSharpProject) -> Project:
        self.default_trill = vspx_project.project.default_trill
        time_signatures = self.parse_time_signatures(vspx_project.project.beat)
        self.first_bar_length = round(time_signatures[0].bar_length())
        tempos = self.parse_tempos(vspx_project.project.tempo)
        self.synchronizer = TimeSynchronizer(tempos)
        singing_tracks = self.parse_singing_tracks(
            [
                track
                for track in vspx_project.project.tracks
                if isinstance(track, VocalSharpNoteTrack)
            ]
        )
        instrumental_tracks = self.parse_instrumental_tracks(
            [
                track
                for track in vspx_project.project.tracks
                if isinstance(track, VocalSharpMonoTrack | VocalSharpStereoTrack)
            ]
        )
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=singing_tracks + instrumental_tracks,
        )

    def parse_time_signatures(self, beat_list: list[VocalSharpBeat]) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=beat.bar_index,
                numerator=beat.beat_per_bar,
                denominator=beat.bar_divide,
            )
            for beat in beat_list
        ]

    def parse_tempos(self, tempo_list: list[VocalSharpTempo]) -> list[SongTempo]:
        return shift_tempo_list(
            [
                SongTempo(
                    position=tempo.pos,
                    bpm=tempo.bpm,
                )
                for tempo in tempo_list
            ],
            self.first_bar_length,
        )

    def parse_singing_tracks(self, track_list: list[VocalSharpNoteTrack]) -> list[SingingTrack]:
        tracks = []
        for track in track_list:
            singing_track = SingingTrack(
                title=track.name,
                mute=track.is_mute == "True",
                solo=track.is_solo == "True",
                ai_singer_name=track.singer,
                note_list=self.parse_notes(track.note),
            )
            if self.options.import_pitch:
                singing_track.edited_params.pitch = self.parse_pitch(track)
            tracks.append(singing_track)
        return tracks

    def parse_notes(self, note_list: list[VocalSharpNote]) -> list[Note]:
        return [
            Note(
                start_pos=note.pos,
                length=note.duration,
                key_number=note.key_number,
                lyric=note.lyric,
                pronunciation=note.lyric,
            )
            for note in note_list
        ]

    def parse_pitch(self, note_track: VocalSharpNoteTrack) -> ParamCurve:
        base_pitch_curve = BasePitchCurve(note_track, self.default_trill, self.synchronizer)
        pitch_points = [Point.start_point()]
        prev_tick = None
        for vspx_point in note_track.parameter.points:
            if isinstance(vspx_point, PIT):
                cur_tick = vspx_point.time + self.first_bar_length
                if prev_tick is None:
                    pitch_points.append(Point(x=cur_tick, y=-100))
                elif vspx_point.time - prev_tick > 1:
                    pitch_points.append(Point(x=prev_tick, y=-100))
                    pitch_points.append(Point(x=cur_tick, y=-100))
                vspx_point_secs = self.synchronizer.get_actual_secs_from_ticks(vspx_point.time)
                if (base_key := base_pitch_curve.semitone_value_at(vspx_point_secs)) is not None:
                    pitch_points.append(
                        Point(
                            x=cur_tick,
                            y=round(base_key * 100 + vspx_point.value),
                        )
                    )
                prev_tick = cur_tick
        if len(pitch_points) > 1:
            pitch_points.append(Point(x=pitch_points[-1].x, y=-100))
        pitch_points.append(Point.end_point())
        if len(pitch_points) <= 2:
            pitch_points.clear()
        return ParamCurve(points=Points(root=pitch_points))

    def parse_instrumental_tracks(
        self,
        track_list: list[VocalSharpMonoTrack | VocalSharpStereoTrack],
    ) -> list[InstrumentalTrack]:
        tracks = []
        if self.options.import_instrumental_track:
            for track in track_list:
                for i, sequence in enumerate(track.sequences):
                    instrumental_track = InstrumentalTrack(
                        title=sequence.name or f"{track.name} {i + 1}",
                        mute=track.is_mute == "True",
                        solo=track.is_solo == "True",
                        offset=sequence.pos,
                        audio_file_path=sequence.path,
                    )
                    tracks.append(instrumental_track)
        return tracks



import enum
from typing import Literal

from pydantic import (
    Field,
    RootModel,
    SerializationInfo,
    ValidationInfo,
    field_serializer,
    field_validator,
)

from libresvip.model.base import BaseModel


class GlideStyle(enum.Enum):
    NONE = "none"
    UP = "up"
    DOWN = "down"


class DsItem(BaseModel):
    text: list[str]
    ph_seq: list[str]
    note_seq: list[str]
    note_dur: list[float] | None = None
    note_dur_seq: list[float] | None = None
    note_slur: list[int] | None = None
    note_glide: list[GlideStyle] | None = None
    is_slur_seq: list[int] | None = None
    ph_dur: list[float] | None = None
    ph_num: list[int] | None = None
    f0_timestep: float | None = None
    f0_seq: str | list[float] | None = None
    input_type: Literal["phoneme"] | None = None
    offset: str | float
    seed: int | None = None
    spk_mix: dict[str, list[float]] | None = None
    spk_mix_timestep: float | None = None
    gender: list[float] | None = None
    gender_timestep: float | None = None
    velocity: list[float] | None = None
    velocity_timestep: float | None = None

    @field_validator("text", "note_seq", "ph_seq", mode="before")
    @classmethod
    def _validate_str_list(cls, value: str | None, _info: ValidationInfo) -> list[str] | None:
        return None if value is None else value.split()

    @field_validator("note_glide", mode="before")
    @classmethod
    def _validate_glide_list(
        cls, value: str | None, _info: ValidationInfo
    ) -> list[GlideStyle] | None:
        return None if value is None else [GlideStyle(x) for x in value.split()]

    @field_validator(
        "f0_seq",
        "ph_dur",
        "note_dur",
        "note_dur_seq",
        "gender",
        "velocity",
        mode="before",
    )
    @classmethod
    def _validate_float_list(cls, value: str | None, _info: ValidationInfo) -> list[float] | None:
        return None if value is None else [float(x) for x in value.split()]

    @field_validator("is_slur_seq", "note_slur", "ph_num", mode="before")
    @classmethod
    def _validate_int_list(cls, value: str | None, _info: ValidationInfo) -> list[int] | None:
        return None if value is None else [int(x) for x in value.split()]

    @field_serializer(
        "f0_seq",
        "ph_num",
        "ph_dur",
        "note_slur",
        "note_dur",
        "note_dur_seq",
        "is_slur_seq",
        "text",
        "note_seq",
        "ph_seq",
        "gender",
        "velocity",
        when_used="json-unless-none",
    )
    @classmethod
    def _serialize_list(cls, value: list[str | int | float], _info: SerializationInfo) -> str:
        return " ".join(str(x) for x in value)

    @field_serializer(
        "note_glide",
        when_used="json-unless-none",
    )
    @classmethod
    def _serialize_glide(cls, value: list[GlideStyle], _info: SerializationInfo) -> str:
        return " ".join(x.value for x in value)

    @field_validator("spk_mix", mode="before")
    @classmethod
    def _validate_nested_dict(
        cls, values: dict[str, str] | None, _info: ValidationInfo
    ) -> dict[str, list[float]] | None:
        if values is None:
            return None

        return {key: [float(x) for x in value.split()] for key, value in values.items()}

    @field_serializer("spk_mix", when_used="json-unless-none")
    @classmethod
    def _serialize_nested_dict(
        cls, value: dict[str, list[float]], _info: SerializationInfo
    ) -> dict[str, str]:
        return {key: " ".join(str(x) for x in value[key]) for key in value}


class DsProject(RootModel[list[DsItem]]):
    root: list[DsItem] = Field(default_factory=list)



# mypy: disable-error-code="misc"
from enum import Enum, IntEnum
from typing import Annotated, NamedTuple

from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableStrengthImportationMixin,
    EnableVolumeImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class SynthVLanguagePreset(NamedTuple):
    language: str
    phoneset: str


synthv_language_presets = {
    "mandarin": SynthVLanguagePreset(language="mandarin", phoneset="xsampa"),
    "cantonese": SynthVLanguagePreset(language="cantonese", phoneset="xsampa"),
    "japanese": SynthVLanguagePreset(language="japanese", phoneset="romaji"),
    "english": SynthVLanguagePreset(language="english", phoneset="arpabet"),
    "spanish": SynthVLanguagePreset(language="spanish", phoneset="xsampa"),
    "korean": SynthVLanguagePreset(language="korean", phoneset="xsampa"),
}


class SVProjectVersionCompatibility(IntEnum):
    BELOW_1_9_0: Annotated[
        int,
        Field(title=_("Compatible with Synthesizer V Studio 1.9.0 and below")),
    ] = 100
    BETWEEN_1_10_0_AND_1_11_2: Annotated[
        int,
        Field(title=_("Compatible with Synthesizer V Studio 1.10.0 to 1.11.2")),
    ] = 135
    ABOVE_2_0_0: Annotated[
        int,
        Field(title=_("Compatible with Synthesizer V Studio 2.0 and up")),
    ] = 182


class LanguageOption(Enum):
    MANDARIN: Annotated[str, Field(title=_("Mandarin"))] = "mandarin"
    CANTONESE: Annotated[str, Field(title=_("Cantonese"))] = "cantonese"
    JAPANESE: Annotated[str, Field(title=_("Japanese"))] = "japanese"
    ENGLISH: Annotated[str, Field(title=_("English"))] = "english"
    SPANISH: Annotated[str, Field(title=_("Spanish"))] = "spanish"
    KOREAN: Annotated[str, Field(title=_("Korean"))] = "korean"


class BreathOption(Enum):
    IGNORE: Annotated[str, Field(title=_("Ignore all breath notes"))] = "ignore"
    KEEP: Annotated[str, Field(title=_("Keep as normal notes"))] = "keep"
    CONVERT: Annotated[str, Field(title=_("Convert to breath mark"))] = "convert"


class GroupOption(Enum):
    SPLIT: Annotated[
        str,
        Field(
            title=_("Split all to tracks"),
            description=_("Generate a track for each note group reference"),
        ),
    ] = "split"
    MERGE: Annotated[
        str,
        Field(
            title=_("Keep original position"),
            description=_("Split note groups to separate tracks only when notes overlap"),
        ),
    ] = "merge"


class PitchOption(Enum):
    FULL: Annotated[
        str,
        Field(
            title=_("Full pitch curve"),
            description=_("Input the full pitch curve regardless of editing"),
        ),
    ] = "full"
    VIBRATO: Annotated[
        str,
        Field(
            title=_("Edited part only (vibrato mode)"),
            description=_(
                "Input the edited part of pitch curve; default vibrato will be imported if not edited"
            ),
        ),
    ] = "vibrato"
    PLAIN: Annotated[
        str,
        Field(
            title=_("Edited part only (plain mode)"),
            description=_("Input the edited part of pitch curve; default vibrato will be ignored"),
        ),
    ] = "plain"


class VibratoOption(Enum):
    NONE: Annotated[
        str,
        Field(
            title=_("All removed"),
            description=_(
                "All notes will be set to 0 vibrato depth to ensure the output pitch curve is the same as input"
            ),
        ),
    ] = "none"
    ALWAYS: Annotated[
        str,
        Field(
            title=_("All kept"),
            description=_(
                "Keep all notes' default vibrato, but may cause inconsistent pitch curves"
            ),
        ),
    ] = "always"
    HYBRID: Annotated[
        str,
        Field(
            title=_("Hybrid mode"),
            description=_("Remove vibrato in edited part, keep default vibrato in other parts"),
        ),
    ] = "hybrid"


class InputOptions(
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableStrengthImportationMixin,
    EnableVolumeImportationMixin,
    BaseModel,
):
    instant: bool = Field(
        default=True,
        title=_("Always follow instant pitch mode"),
        description=_(
            "When this option is turned off, the default pitch curve will always be imported regardless of the project setting. If you have tuned the pitch curve based on instant pitch mode, it is recommended to turn on this option."
        ),
    )
    pitch: PitchOption = Field(
        default=PitchOption.PLAIN,
        title=_("Pitch input mode"),
        description=_(
            'This option controls the range of pitch curve to be imported and the judgment condition. The definition of "edited part" is: the pitch deviation in the parameter panel, the pitch transition in the vibrato envelope and the pitch transition in the note properties have been edited.'
        ),
    )
    breath: BreathOption = Field(
        default=BreathOption.CONVERT,
        title=_("The way to handle breath notes"),
    )
    group: GroupOption = Field(
        default=GroupOption.SPLIT,
        title=_("The way to handle note groups"),
        description=_(
            'Notice: If there are too many note groups, please choose "Keep original position" to avoid excessive track count. But if there are notes that are adjacent (but not overlapped) between note groups or between note groups and main group, it is recommended to choose "Split to tracks" to ensure the paragraph division is not broken.'
        ),
    )


class OutputOptions(BaseModel):
    version_compatibility: SVProjectVersionCompatibility = Field(
        default=SVProjectVersionCompatibility.BELOW_1_9_0,
        title=_("Version compatibility"),
    )
    vibrato: VibratoOption = Field(
        default=VibratoOption.NONE, title=_("The way to handle vibrato notes")
    )
    down_sample: int = Field(
        default=40,
        title=_(
            "Set the average sampling interval of parameter points to improve performance (0 means no limit)"
        ),
        description=_(
            "Reduce the sampling interval to improve the accuracy of parameter curves, but may cause rendering lag (e.g. Synthesizer V Studio Pro + AI voicebank). Please set this value according to your hardware configuration and actual experience."
        ),
    )
    language_override: LanguageOption = Field(
        default=LanguageOption.MANDARIN,
        title=_("Override default language for the voicebank"),
    )



import math


class LambertW:
    @classmethod
    def evaluate(cls, x: float, branch: int = 0) -> float:
        if x < -1 / math.e:
            msg = "x must be greater than -1/e"
            raise ValueError(msg)
        if branch not in (0, -1):
            msg = "branch must be 0 or -1"
            raise ValueError(msg)
        if branch == -1 and x > 0:
            msg = "x must be less than 0 for branch -1"
            raise ValueError(msg)
        result = cls.estimate(x, branch)
        loop_remaining = 100
        while cls.test_result(result, x) and loop_remaining > 0:
            x0 = result
            y0 = result * math.exp(result)
            k = math.exp(x0) * (1 + x0)
            x1 = (x - y0) / k + x0
            result = x1
            loop_remaining -= 1
        if loop_remaining == 0:
            msg = "Cannot find solution"
            raise ValueError(msg)
        return result

    @classmethod
    def estimate(cls, x: float, branch: int = 0) -> float:
        if branch == -1:
            return -2 if x > -0.2706706 else -1 - math.acos(-4.11727 * x - 0.514659)
        else:
            return -1 + math.log(1 / math.e + math.e + x)

    @classmethod
    def test_result(cls, x: float, y: float) -> bool:
        fx = x * math.exp(x)
        if (delta := fx - y) == 0:
            return False
        delta_bits = math.log2(abs(y / delta))
        return delta_bits < 42



import dataclasses
import itertools

import more_itertools
from drawsvg import Line, Lines, Rectangle, Text

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, TimeSignature
from libresvip.model.point import Point

from .coordinate_helper import CoordinateHelper
from .model import NotePositionParameters
from .options import OutputOptions, TextPositionOption


@dataclasses.dataclass
class SvgFactory:
    coordinate_helper: CoordinateHelper
    time_synchronizer: TimeSynchronizer
    options: OutputOptions
    line_elements: list[Line] = dataclasses.field(default_factory=list)
    polyline_elements: list[Lines] = dataclasses.field(default_factory=list)
    rect_elements: list[Rectangle] = dataclasses.field(default_factory=list)
    text_elements: list[Text] = dataclasses.field(default_factory=list)
    pitch_points_buf: list[Point] = dataclasses.field(default_factory=list)
    style: str = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.style = f"""\
.note {{
    fill: {self.options.note_fill_color.as_hex()};
    stroke: {self.options.note_stroke_color.as_hex()};
    stroke-width: {self.options.note_stroke_width}px;
}}
.pitch {{
    fill: none;
    stroke: {self.options.pitch_stroke_color.as_hex()};
    stroke-width: {self.options.pitch_stroke_width}px;
}}
text {{
    font-size: {self.coordinate_helper.font_size}px;
    text-anchor: {self.coordinate_helper.text_anchor};
}}
.inner {{
    fill: {self.options.inner_text_color.as_hex()};
}}
.side {{
    fill: {self.options.side_text_color.as_hex()};
    font-size: {max(self.coordinate_helper.font_size - 4, 10)}px;
}}
.line {{
    fill: none;
    stroke: {self.options.grid_color.as_hex()};
    stroke-width: {self.options.grid_stroke_width}px;
}}"""

    def draw_grid(self, time_signature_list: list[TimeSignature]) -> None:
        prev_pos = 0
        pos = 0
        beat_start = self.coordinate_helper.position_range_start
        for previous_time_signature, time_signature in more_itertools.pairwise(time_signature_list):
            beat_length = (
                time_signature.bar_index - previous_time_signature.bar_index
            ) * previous_time_signature.numerator
            for beat_offset in range(beat_length):
                pos = prev_pos + beat_offset * TICKS_IN_BEAT
                if pos >= beat_start:
                    x = int(
                        (pos - self.coordinate_helper.position_range_start)
                        * self.options.pixel_per_beat
                        / TICKS_IN_BEAT
                    )
                    line_element = Line(
                        sx=x,
                        sy=0,
                        ex=x,
                        ey=self.coordinate_helper.size[1],
                        class_="line",
                    )
                    self.line_elements.append(line_element)
            prev_pos += beat_length * TICKS_IN_BEAT
        if len(time_signature_list):
            pos = prev_pos
            while pos <= self.coordinate_helper.position_range_end:
                pos += TICKS_IN_BEAT
                if pos >= beat_start:
                    x = int((pos - beat_start) * self.options.pixel_per_beat / TICKS_IN_BEAT)
                    line_element = Line(
                        sx=x,
                        sy=0,
                        ex=x,
                        ey=self.coordinate_helper.size[1],
                        class_="line",
                    )
                    self.line_elements.append(line_element)
        for key in range(
            self.coordinate_helper.key_range_start,
            self.coordinate_helper.key_range_end + 1,
        ):
            y = (self.coordinate_helper.key_range_end - key) * self.options.note_height
            line_element = Line(
                sx=0,
                sy=y,
                ex=self.coordinate_helper.size[0],
                ey=y,
                class_="line",
            )
            self.line_elements.append(line_element)

    def draw_text(
        self,
        position: TextPositionOption,
        text: str | None,
        parameters: NotePositionParameters,
        is_phoneme: bool,
    ) -> None:
        if not text:
            return
        elif position == TextPositionOption.UPPER:
            insert_pos = parameters.upper_text
        elif position == TextPositionOption.LOWER:
            insert_pos = parameters.lower_text
        elif position == TextPositionOption.INNER:
            insert_pos = parameters.inner_text
        else:
            return
        text_element = Text(
            text,
            self.coordinate_helper.font_size,
            x=insert_pos[0],
            y=insert_pos[1],
        )
        class_name = "inner" if position == TextPositionOption.INNER else "side"
        if is_phoneme:
            class_name += " pinyin"
        text_element.args["class"] = class_name
        self.text_elements.append(text_element)

    def draw_note(self, note: Note) -> None:
        parameters = self.coordinate_helper.get_note_position_parameters(note)
        rect_element = Rectangle(
            x=parameters.point_1[0],
            y=parameters.point_1[1],
            width=parameters.point_2[0] - parameters.point_1[0],
            height=parameters.point_2[1] - parameters.point_1[1],
            rx=self.options.note_round,
            ry=self.options.note_round,
            class_="note",
        )
        self.rect_elements.append(rect_element)
        self.draw_text(self.options.lyric_position, note.lyric, parameters, False)
        self.draw_text(
            self.options.pronounciation_position,
            note.pronunciation,
            parameters,
            True,
        )

    def draw_pitch(self, point: Point) -> None:
        if point.y == -100:
            if len(self.pitch_points_buf):
                start_point = self.coordinate_helper.get_pitch_point(self.pitch_points_buf[0])
                polyline_element = Lines(
                    start_point[0],
                    start_point[1],
                    *itertools.chain.from_iterable(
                        self.coordinate_helper.get_pitch_point(p) for p in self.pitch_points_buf[1:]
                    ),
                    class_="pitch",
                )
                self.polyline_elements.append(polyline_element)
                self.pitch_points_buf.clear()
        else:
            self.pitch_points_buf.append(point)



from libresvip.model.base import ParamCurve, Points
from libresvip.model.point import Point
from libresvip.utils.search import binary_find_first, binary_find_last

from .models.ds_param_curve import DsParamCurve
from .models.ds_param_node import DsParamNode


class GenderParamUtils:
    @classmethod
    def encode(cls, curve: ParamCurve, end: int, time_step: float = 0.005) -> DsParamCurve:
        end += 1920
        return DsParamCurve(
            step_size=time_step,
            point_list=cls.encode_point_list(curve.points, end),
        )

    @classmethod
    def encode_point_list(cls, os_point_list: Points, end: int) -> list[DsParamNode]:
        valid_points = [
            p for p in os_point_list.root if p.x >= 1930 and p.x + 10 < end and -1000 <= p.y <= 1000
        ]
        return [
            DsParamNode(
                time=pos / 1000.0,
                value=cls.value_at(valid_points, pos) / 1000.0,
            )
            for pos in range(1920, end, 5)
        ]

    @staticmethod
    def value_at(segment: list[Point], ticks: float) -> float:
        left_point = binary_find_last(segment, lambda point: point.x <= ticks)
        if left_point is None:
            return segment[0].y

        right_point = binary_find_first(segment, lambda point: point.x > ticks)
        if right_point is None:
            return segment[-1].y

        ratio = (ticks - left_point.x) / (right_point.x - left_point.x)
        return (1 - ratio) * left_point.y + ratio * right_point.y



import io
import pathlib

from libresvip.core.compat import ZipFile, json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis

from .model import VogenProject
from .options import InputOptions, OutputOptions
from .vogen_generator import VogenGenerator
from .vogen_parser import VogenParser


class VogenConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        with ZipFile(io.BytesIO(path.read_bytes()), "r") as zf:
            proj_text = zf.read("chart.json")
        vogen_project = VogenProject.model_validate_json(proj_text)
        return VogenParser(options).parse_project(vogen_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        if len(project.song_tempo_list) != 1:
            project = reset_time_axis(project, options.tempo)
        vogen_project = VogenGenerator(options).generate_project(project)
        proj_text = json.dumps(vogen_project.model_dump(by_alias=True), separators=(",", ":"))
        buffer = io.BytesIO()
        with ZipFile(buffer, "w") as zf:
            zf.writestr("chart.json", proj_text)
        path.write_bytes(buffer.getvalue())



import dataclasses
import struct
from collections import defaultdict

from construct import (
    Byte,
    Container,
    Int16ul,
    PascalString,
    PrefixedArray,
    Struct,
)
from more_itertools import split_into

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)

from .legacy_model import PpsfChunk, PpsfLegacyProject
from .options import InputOptions


@dataclasses.dataclass
class PiaproStudioLegacyParser:
    options: InputOptions
    clip_offsets: list[int] = dataclasses.field(default_factory=list)
    clip_note_counts: list[int] = dataclasses.field(default_factory=list)
    clips2track_indexes: dict[int, int] = dataclasses.field(default_factory=dict)

    def parse_project(self, ppsf_project: PpsfLegacyProject) -> Project:
        events_chunk: PpsfChunk | None = None
        for chunk in ppsf_project.body.chunks:
            match chunk.magic:
                case "Clips":
                    for clip_group in chunk.data.clips:
                        for clip in clip_group:
                            if clip.magic == "Vocaloid3NoteClip":
                                (clip_offset,) = struct.unpack_from(
                                    "<i", clip.data, len(clip.data) - 32
                                )
                                self.clip_offsets.append(clip_offset)
                case "EditorDatas":
                    for track_data in chunk.data.editor_datas:
                        for clip_data in track_data.data.clip_datas:
                            self.clip_note_counts.append(len(clip_data.data.note_datas))
                case "Events":
                    events_chunk = chunk
                case "Tracks":
                    clip_indexes_struct = Struct(
                        PascalString(Byte, "utf-8"),
                        "indexes" / PrefixedArray(Byte, Int16ul),
                    )
                    for track_group in chunk.data.tracks:
                        if track_group.magic == "Vocaloid3EventTracks":
                            clip_index = 0
                            for i, track in enumerate(track_group.data):
                                clip_indexes = clip_indexes_struct.parse(track.data[30:])
                                for clip_index in clip_indexes.indexes:
                                    self.clips2track_indexes[clip_index] = i
        tempos, time_signatures = self.parse_tempos_and_time_signatures(events_chunk)
        return Project(
            song_tempo_list=tempos,
            time_signature_list=time_signatures,
            track_list=self.parse_tracks(events_chunk),
        )

    def parse_tempos_and_time_signatures(
        self, events_chunk: Container | None = None
    ) -> tuple[list[SongTempo], list[TimeSignature]]:
        tempos = []
        time_signatures: list[TimeSignature] = []
        if events_chunk is not None:
            prev_tick = 0
            for event_group in events_chunk.data.events:
                events_by_level: list[list[tuple[int, bytes]]] = []
                for event in event_group:
                    if event.magic == "MidiEvent":
                        (tick,) = struct.unpack_from(
                            "<i",
                            event.data,
                        )
                        if event.data[-1] == 0:
                            if tick == 0:
                                events_by_level.append([])
                            if events_by_level:
                                events_by_level[-1].append((tick, event.data))

                if len(events_by_level) > 1:
                    for tick, event_data in events_by_level[-2]:
                        bpm = struct.unpack_from("<i", event_data, 4)[0] / 10000
                        tempos.append(
                            SongTempo(
                                position=tick,
                                bpm=bpm,
                            )
                        )
                    for tick, event_data in events_by_level[-1]:
                        (denominator, numerator) = struct.unpack_from("<2b", event_data, 4)
                        if time_signatures:
                            prev_bar_length = time_signatures[-1].bar_length()
                            prev_bar_index = time_signatures[-1].bar_index
                        else:
                            prev_bar_length = TICKS_IN_BEAT * 4
                            prev_bar_index = 0
                        time_signatures.append(
                            TimeSignature(
                                bar_index=prev_bar_index + (tick - prev_tick) // prev_bar_length,
                                numerator=numerator,
                                denominator=denominator,
                            )
                        )
                        prev_tick = tick
        if not tempos:
            tempos.append(SongTempo())
        if not time_signatures:
            time_signatures.append(TimeSignature())
        return tempos, time_signatures

    def parse_tracks(self, events_chunk: Container | None = None) -> list[SingingTrack]:
        tracks: list[SingingTrack] = []
        if events_chunk is not None:
            lyric_info_struct = Struct(
                "lyric" / PascalString(Byte, "utf-8"),
                Byte,
                "phoneme" / PascalString(Byte, "utf-8"),
                Byte,
            )
            track_index2notes = defaultdict(list)
            note_events = [
                event
                for event_group in events_chunk.data.events
                for event in event_group
                if event.magic == "Vocaloid3NoteEvent"
            ]
            for i, note_group in enumerate(split_into(note_events, self.clip_note_counts)):
                for note in note_group:
                    note_offset, pit, length = struct.unpack_from(
                        "<ibi",
                        note.data,
                    )
                    lyric_info = lyric_info_struct.parse(note.data[16:])
                    track_index2notes[self.clips2track_indexes[i]].append(
                        Note(
                            start_pos=self.clip_offsets[i] + note_offset,
                            length=length,
                            key_number=pit,
                            lyric=lyric_info.lyric,
                            pronunciation=lyric_info.phoneme,
                        )
                    )
            tracks.extend(SingingTrack(note_list=notes) for notes in track_index2notes.values())
        return tracks



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import Field

from libresvip.utils.translation import gettext_lazy as _


class AcepLyricsLanguage(Enum):
    CHINESE: Annotated[
        str,
        Field(
            title=_("Chinese"),
        ),
    ] = "CHN"
    JAPANESE: Annotated[
        str,
        Field(
            title=_("Japanese"),
        ),
    ] = "JPN"
    ENGLISH: Annotated[
        str,
        Field(
            title=_("English"),
        ),
    ] = "ENG"
    SPANISH: Annotated[
        str,
        Field(
            title=_("Spanish"),
        ),
    ] = "SPA"



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import muta_project_struct
from .muta_generator import MutaGenerator
from .muta_parser import MutaParser
from .options import InputOptions, OutputOptions


class MutaConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        muta_project = muta_project_struct.parse(path.read_bytes())
        return MutaParser(options).parse_project(muta_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        muta_project = MutaGenerator(options).generate_project(project)
        path.write_bytes(muta_project_struct.build(muta_project))



from __future__ import annotations

from itertools import chain
from typing import Annotated, Any, Literal, NamedTuple, cast

from more_itertools import batched
from pydantic import (
    Field,
    RootModel,
    ValidationInfo,
    field_validator,
    model_serializer,
)

from libresvip.model.base import BaseModel
from libresvip.model.point import PointList


class TuneLabPoint(NamedTuple):
    pos: float
    value: float


class TuneLabPoints(PointList[TuneLabPoint], RootModel[list[TuneLabPoint]]):
    root: list[TuneLabPoint] = Field(default_factory=list)

    @model_serializer(when_used="json")
    def serialize_points(self) -> list[float]:
        return list(chain.from_iterable(self.root))


class TuneLabTempo(BaseModel):
    pos: float
    bpm: float


class TuneLabTimeSignature(BaseModel):
    numerator: int
    denominator: int
    bar_index: int = Field(alias="barIndex")


class TuneLabVoice(BaseModel):
    type_: str = Field("", alias="type")
    id_: str = Field("", alias="id")


class TuneLabAutomation(BaseModel):
    default: float
    values: TuneLabPoints = Field(default_factory=TuneLabPoints)

    @field_validator("values", mode="before")
    @classmethod
    def validate_values(
        cls, values: list[float | TuneLabPoint], _info: ValidationInfo
    ) -> TuneLabPoints:
        if _info.mode == "json":
            return TuneLabPoints(root=[TuneLabPoint._make(each) for each in batched(values, 2)])
        return TuneLabPoints(root=values)


class TuneLabPhoneme(BaseModel):
    start_time: float = Field(alias="startTime")
    end_time: float = Field(alias="endTime")
    symbol: str


class TuneLabNote(BaseModel):
    pos: float
    dur: float
    pitch: int
    lyric: str
    pronunciation: str | None = None
    phonemes: list[TuneLabPhoneme] = Field(default_factory=list)
    properties: dict[str, Any] = Field(default_factory=dict)


class TuneLabVibrato(BaseModel):
    pos: float
    dur: float
    frequency: float
    amplitude: float
    phase: float
    attack: float
    release: float
    affected_automations: dict[str, float] = Field(
        default_factory=dict, alias="affectedAutomations"
    )


class TuneLabBasePart(BaseModel):
    name: str
    pos: float
    dur: float


class TuneLabMidiPart(TuneLabBasePart):
    type_: Literal["midi"] = Field("midi", alias="type")
    gain: float | None = 0.0
    voice: TuneLabVoice = Field(default_factory=TuneLabVoice)
    properties: dict[str, Any] = Field(default_factory=dict)
    notes: list[TuneLabNote] = Field(default_factory=list)
    automations: dict[str, TuneLabAutomation] = Field(default_factory=dict)
    pitch: list[TuneLabPoints] = Field(default_factory=list)
    vibratos: list[TuneLabVibrato] = Field(default_factory=list)

    @field_validator("pitch", mode="before")
    @classmethod
    def validate_pitch(
        cls,
        pitch: list[list[float]] | list[TuneLabPoints],
        _info: ValidationInfo,
    ) -> list[TuneLabPoints]:
        if _info.mode == "json":
            return [
                TuneLabPoints(root=[TuneLabPoint._make(each) for each in batched(values, 2)])
                for values in pitch
            ]
        return cast("list[TuneLabPoints]", pitch)


class TuneLabAudioPart(TuneLabBasePart):
    type_: Literal["audio"] = Field("audio", alias="type")
    path: str


TuneLabPart = Annotated[TuneLabMidiPart | TuneLabAudioPart, Field(discriminator="type_")]


class TuneLabTrack(BaseModel):
    name: str
    gain: float
    pan: float
    mute: bool
    solo: bool
    color: str = ""
    as_refer: bool | None = Field(False, alias="asRefer")
    parts: list[TuneLabPart] = Field(default_factory=list)


class TuneLabProject(BaseModel):
    version: int = 0
    tempos: list[TuneLabTempo] = Field(default_factory=list)
    time_signatures: list[TuneLabTimeSignature] = Field(
        default_factory=list, alias="timeSignatures"
    )
    tracks: list[TuneLabTrack] = Field(default_factory=list)
